#!/usr/bin/env python

############################################################################
# Copyright (c) 2013 Sebastian Clarke, RSGISLib
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#
# Purpose:  Test suite for RSGISLib python bindings
# Author: Sebastian Clarke
# Email: sec14@aber.ac.uk
# Date: 30/07/2013
# Version: 1.0
# Version: 1.1 - Data copying now has Windows support.
# Version: 1.2 - Updated to use argparse
#############################################################################

import sys, os
import shutil
import argparse
import collections
try:
    import rsgislib
    from rsgislib import imageutils
    from rsgislib import vectorutils
    from rsgislib import imagecalc
    from rsgislib import rastergis
    from rsgislib import zonalstats
    from rsgislib import imageregistration
    from rsgislib import imagefilter
    from rsgislib import segmentation
    from rsgislib.segmentation import segutils
    from rsgislib.imagecalc import BandDefn
    from rsgislib import tools
except ImportError as err:
    print(err)
    sys.exit()
    

path = os.sys.path[0] + '/'
inFileName = os.path.join(path,"Rasters","injune_p142_casi_sub_utm.kea")
print(inFileName)

class RSGISTests:

    def __init__(self):
        self.numTests = 0
        self.failures = []
        self.testOutputsDIR = os.path.join(path,'TestOutputs')
        self.testRasterGISDIR = os.path.join(path,'TestOutputs','RasterGIS')
        self.testTilesDIR = os.path.join(path,'TestOutputs','Tiles')
        self.testZonalTXTDIR = os.path.join(path,'TestOutputs','ZonalTXT')

    def tryFuncAndCatch(self, function):
        self.numTests += 1
        try:
            function()
        except Exception as inst:
            print("ERROR - ERROR - ERROR")
            print("inst")
            self.failures.append(function.__name__)

    def copyData(self):
        """ Copy data files from original directory to test directory """
        print("COPYING DATA")
        shutil.copy2('RATS/injune_p142_casi_sub_utm_segs_nostats.kea', 'TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_cptab.kea')
        shutil.copy2('RATS/injune_p142_casi_sub_utm_segs_nostats.kea', 'TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_cpcols.kea')
        shutil.copy2('RATS/injune_p142_casi_sub_utm_segs_nostats.kea', 'TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_spatloc_eucdist.kea')
        shutil.copy2('RATS/injune_p142_casi_sub_utm_segs_nostats.kea', 'TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_popstats.kea')
        shutil.copy2('RATS/injune_p142_casi_sub_utm_segs.kea', 'TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_neighbours.kea')
        shutil.copy2('RATS/injune_p142_casi_sub_utm_segs.kea', 'TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_borlen.kea')
        shutil.copy2('RATS/injune_p142_casi_sub_utm_segs.kea', 'TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_borlen.kea')
        shutil.copy2('RATS/injune_p142_casi_sub_utm_segs.kea', 'TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_shape.kea')
        shutil.copy2('RATS/injune_p142_casi_sub_utm_segs.kea', 'TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_col.kea')
        shutil.copy2('RATS/injune_p142_casi_sub_utm_segs.kea', 'TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_col_str.kea')
        shutil.copy2('RATS/injune_p142_casi_sub_utm_segs.kea', 'TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_change.kea')
        shutil.copy2('Rasters/injune_p142_casi_sub_utm.kea', 'TestOutputs/injune_p142_casi_sub_utm.kea')
        
        shutil.copy2('RATS/injune_p142_casi_sub_utm_segs_nostats.kea', 'TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_nostats_addstats.kea')
        shutil.copy2('Vectors/injune_p142_psu_utm.dbf', 'TestOutputs/')
        shutil.copy2('Vectors/injune_p142_psu_utm.prj', 'TestOutputs/')
        shutil.copy2('Vectors/injune_p142_psu_utm.sbn', 'TestOutputs/')
        shutil.copy2('Vectors/injune_p142_psu_utm.sbx', 'TestOutputs/')
        shutil.copy2('Vectors/injune_p142_psu_utm.shp', 'TestOutputs/')
        shutil.copy2('Vectors/injune_p142_psu_utm.shp_utm.xml', 'TestOutputs/')
        shutil.copy2('Vectors/injune_p142_psu_utm.shx', 'TestOutputs/')
    
    def checkDIRStructure(self):
        """ Create directory and sub directories for test
            outputs, if they don't already exist.  """
        if os.path.isdir(self.testOutputsDIR) == False:
            os.mkdir(self.testOutputsDIR)

        if os.path.isdir(self.testRasterGISDIR) == False:
            os.mkdir(self.testRasterGISDIR)
            
        if os.path.isdir(self.testTilesDIR) == False:
            os.mkdir(self.testTilesDIR)
    
        if os.path.isdir(self.testZonalTXTDIR) == False:
            os.mkdir(self.testZonalTXTDIR)

    def removeTestFiles(self):
        """ Removes all files in test directory """
        print('Removing test files')
        shutil.rmtree('TestOutputs/')

    # Image Calc

    def testNormalise1(self):
        print("PYTHON TEST: Testing normalisation no in calc")
        inImages = [inFileName];
        outImages = [path+"TestOutputs/injune_p142_casi_sub_ll_norm.env"]
        imagecalc.normalisation(inImages, outImages, False, 0, 0, 0, 100)

    def testNormalise2(self):
        print("PYTHON TEST: Testing normalisation with in calc")
        inImages = [inFileName];
        outImages = [path+"TestOutputs/injune_p142_casi_sub_ll_norm.env"]
        imagecalc.normalisation(inImages, outImages, True, 0, 1000, 0, 50)

    def testCorrelation(self):
        print("PYTHON TEST: Testing correlation")
        inImageA = inFileName
        inImageB = inFileName
        outMatrix = imagecalc.correlation(inImageA, inImageB)
        if int(outMatrix[0][0]) != 1:
            raise Exception('First value of output correlation matrix is incorrect, expected 1, got {}'.format(outMatrix[0][0]))

    def testCovariance1(self):
        print("PYTHON TEST: Testing covariance - with mean calculation")
        inImage = inFileName
        outImage = path+"TestOutputs/PSU142Correlation"
        imagecalc.covariance(inImage, inImage, "", "", True, outImage)

    def testCovariance2(self):
        # TODO: Missing matrix files
        print("PYTHON TEST: Testing covariance - no mean calculation")
        inImage = inFileName
        outImage = path+"TestOutputs/matrix"
        imagecalc.covariance(inImage, inImage, "matrixA.mtxt", "matrixB.mtxt", False, outImage)

    def testCalcRMSE(self):
        # note erratic behaviour
        print("PYTHON TEST: Testing RMSE")
        inImageA = inFileName
        inImageB = inFileName
        bandA = 1
        bandB = 5
        imagecalc.calculateRMSE(inImageA, bandA, inImageB, bandB)

    def testMeanVector(self):
        print("PYTHON TEST: Testing meanVector")
        inImage = inFileName
        outVector = path+"TestOutputs/PSU142MeanVec"
        imagecalc.meanVector(inImage, outVector)

    def testPCA(self):
        print("PYTHON TEST: unable to test PCA - awaiting eigenvectors")

    def testStandardise(self):
        print("PYTHON TEST: Testing standardise")
        outImage = path + "TestOutputs/PSU142_Standarised.env"
        meanVector = path + "TestOutputs/PSU142MeanVec.mtxt"
        imagecalc.standardise(meanVector, inFileName, outImage)

    def testBandMath(self):
        # NOT WORKING band not in image
        print("PYTHON TEST: Testing bandMath")
        outputImage = path + "TestOutputs/PSU142_b1mb2.kea"
        gdalformat = "KEA"
        dataType = rsgislib.TYPE_32FLOAT
        expression = "b1*b2"
        bandDefns = []
        bandDefns.append(BandDefn("b1", inFileName, 1))
        bandDefns.append(BandDefn("b2", inFileName, 2))
        imagecalc.bandMath(outputImage, expression, gdalformat, dataType, bandDefns)

    def testImageMaths(self):
        print("PYTHON TEST: Testing imageMath")
        outputImage = path + "TestOutputs/PSU142_multi1000.kea"
        gdalformat = "KEA"
        dataType = rsgislib.TYPE_32UINT
        expression = "1e3*b1"
        imagecalc.imageMath(inFileName, outputImage, expression, gdalformat, dataType)

    def testReplaceValuesLessThan(self):
        print("PYTHON TEST: Replace values less than")
        outputImage = path + "TestOutputs/injune_p142_casi_sub_ll_lt100.kea"
        imagecalc.replaceValuesLessThan(inFileName, outputImage, 100.0, 0)

    def testUnitArea(self):
        print("PYTHON TEST: unit area - CANNOT TEST - awaiting imagebands matrix.mtxt")

    def testMovementSpeed(self):
        print("PYTHON TEST: movement speed - CANNOT TEST - awaiting multiple images")

    def testCountValsInCol(self):
        print("PYTHON TEST: countValsInCols")
        outputImage = path + "TestOutputs/injune_p142_casi_sub_ll_count.kea"
        upper = 1000
        lower = 500
        imagecalc.countValsInCols(inFileName, upper, lower, outputImage)

    def testDist2Geoms(self):
        print("PYTHON TEST: dist2Geoms - not testing due to poor time perfomance and lack of example geometries")

    def testImgDist2Geoms(self):
        print("PYTHON TEST: imgDist2Geoms - not testing due to poor time perfomance and lack of example geometries")

    def testImgCalcDist(self):
        print("PYTHON TEST: imgCalcDist - not testing due to no example in XML")

    def testImageBandStats(self):
        print("PYTHON TEST: imageBandStats - not ignoring Zeros")
        outputImage = path + "TestOutputs/BandsStats.txt"
        imagecalc.imageBandStats(inFileName, outputImage, False)

    def testImageBandStatsIgnoreZeros(self):
        print("PYTHON TEST: imageBandStats - Ignoring Zeros")
        outputImage = path + "TestOutputs/BandsStatsIgnore.txt"
        imagecalc.imageBandStats(inFileName, outputImage, True)

    def testImageStats(self):
        print("PYTHON TEST: imageStats - Not Ignoring Zeros")
        outputImage = path + "TestOutputs/ImageStats.txt"
        imagecalc.imageStats(inFileName, outputImage, False)

    def testImageStatsIgnoreZeros(self):
        print("PYTHON TEST: imageStats - Ignoring Zeros")
        outputImage = path + "TestOutputs/ImageStats.txt"
        imagecalc.imageStats(inFileName, outputImage, True)

    def testUnconLinearSpecUnmix(self):
        print("PYTHON TEST: unconLinearSpecUnmix - skipping due to lack of test data")

    def testExhConLinearSpecUnmix(self):
        print("PYTHON TEST: exhConinearSpecUnmix - skipping due to lack of test data")

    def testConSum1LinearSpecUnmix(self):
        print("PYTHON TEST: ConSum1LinearSpecUnmix - skipping due to lack of test data")

    def testNnConSum1LinearSpecUnmix(self):
        print("PYTHON TEST: NnConSum1LinearSpecUnmix - skipping due to lack of test data")

    def testKMeansCentres(self):
        # Only seems to do one iteration?
        print("PYTHON TEST: kMeansClustering")
        inputImage = path + "Rasters/injune_p142_casi_sub_right_utm.kea"
        output = path + "TestOutputs/kmeanscentres"
        numClust = 10
        maxIter = 200
        degChange = 0.0025
        subSample = 1
        ignoreZeros = True
        imagecalc.kMeansClustering(inputImage, output, numClust, maxIter, subSample, ignoreZeros, degChange, rsgislib.INITCLUSTER_DIAGONAL_FULL_ATTACH)

    def testIsoDataClustering(self):
        # Only seems to do one iteration?
        print("PYTHON TEST: isoDataClustering")
        inputImage = path + "Rasters/injune_p142_casi_sub_right_utm.kea"
        output = path + "TestOutputs/isocentres"
        imagecalc.isoDataClustering(inputImage, output, 10, 200, 1, True, 0.0025, rsgislib.INITCLUSTER_DIAGONAL_FULL_ATTACH, 2, 5, 5, 5, 8, 50)

    def testAllBandsEqualTo(self):
        print("PYTHON TEST: allBandsEqualTo")
        inputImage = path + "Rasters/injune_p142_casi_sub_right_utm.kea"
        output = path + "TestOutputs/PSU142_Allbandsequal0Mask.kea"
        gdalformat = "KEA"
        value = 0
        trueout = 0
        falseout = 1
        dataType = rsgislib.TYPE_8INT
        imagecalc.allBandsEqualTo(inputImage, value, trueout, falseout, output, gdalformat, dataType)

    def testHistogram(self):
        print("PYTHON TEST: Histogram")
        image = path + "Rasters/injune_p142_casi_sub_right_utm.kea"
        out = path + "TestOutputs/injune_p142_casi_sub_right_utm_histo.txt"
        mask = path + "TestOutputs/PSU142_Allbandsequal0Mask.kea"
        value = 1
        band = 1
        imagecalc.histogram(image, mask, out, band, value, 5, True, 0, 0)

    def testBandPercentile(self):
        print("PYTHON TEST: Band Percentile")
        image = path + "Rasters/injune_p142_casi_sub_right_utm.kea"
        percentile = 0.25
        nodata = 0
        outpercentiles = imagecalc.bandPercentile(image, percentile, nodata)
        if int(outpercentiles[0]) != 2723:
            raise Exception('Incorrect percentile value returned. Expected 2723, got {}'.format(outpercentiles[0]))

    def testMahalanobisDistWindow(self):
        print("PYTHON TEST: MahalanobisDistWindow")
        image = path + "Rasters/injune_p142_casi_sub_right_utm.kea"
        output = path + "TestOutputs/PSU142_MabDistWin.kea"
        window = 7
        gdalformat = "KEA"
        dataType = rsgislib.TYPE_32FLOAT
        imagecalc.mahalanobisDistFilter(image, output, window, gdalformat, dataType)

    def testMahalanobisDistImg2Window(self):
        print("PYTHON TEST: MahalanobisDistImg2Window")
        image = path + "Rasters/injune_p142_casi_sub_right_utm.kea"
        output = path + "TestOutputs/PSU142_MabDist2ImgWin.kea"
        window = 7
        gdalformat = "KEA"
        dataType = rsgislib.TYPE_32FLOAT
        imagecalc.mahalanobisDist2ImgFilter(image, output, window, gdalformat, dataType)

    def testCalcPxlColStats(self):
        print("PYTHON TEST: pxlColStats")
        image = path + "Rasters/injune_p142_casi_sub_right_utm.kea"
        output = path + "TestOutputs/PSU142_ImageStats.kea"
        gdalformat = "KEA"
        dataType = rsgislib.TYPE_32FLOAT
        s = imagecalc.StatsSummary(calcMin=True, calcMax=True, calcSum=True, calcMean=True, calcMedian=True, calcStdDev=True, min=0.0)
        imagecalc.imagePixelColumnSummary(image, output, s, gdalformat, dataType, 0, True)

    def testCorrelationWindow(self):
        print("PYTHON TEST: correlationWindow")
        image = path + "Rasters/injune_p142_casi_sub_utm.kea"
        output = path + "TestOutputs/injune_p142_casi_sub_utm_correlation.kea"
        window = 9
        bandA = 1
        bandB = 1
        gdalformat = "KEA"
        dataType = rsgislib.TYPE_32FLOAT
        imagecalc.correlationWindow(image, output, window, bandA, bandB, gdalformat, dataType)


    def testImagePixelLinearFit(self):
        print("PYTHON TEST: imagePixelLinearFit")
        image = path + "Rasters/injune_p142_casi_sub_utm.kea"
        output = path + "TestOutputs/injune_p142_casi_sub_utm_linear_fit.kea"
        gdalformat = "KEA"
        bandValues = [446,530,549,569,598,633,680,696,714,732,741,752,800,838]
        
        bandValuesFile =  path + "TestOutputs/injune_p142_casi_wavelengths.txt"
         
        with open(bandValuesFile,'w') as f:
            for bandVal in bandValues:
                f.write(str(bandVal) + '\n')

        imagecalc.imagePixelLinearFit(image, output, gdalformat, bandValuesFile, 0, True)

    # Raster GIS

    def testCopyGDLATT(self):
        print("PYTHON TEST: copyRAT")
        table = "./RATS/injune_p142_casi_sub_utm_segs.kea"
        image = "./TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_cptab.kea"
        rastergis.copyRAT(image, table)

    def testCopyGDLATTColumns(self):
        print("PYTHON TEST: copyGDALATTColumns")
        table = "./RATS/injune_p142_casi_sub_utm_segs.kea"
        image = "./TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_cpcols.kea"
        fields = ["b1Mean", "outClass", "outClassStr"]
        rastergis.copyGDALATTColumns(table, image, fields, copycolours=True, copyhist=False)

    """
    def testSpatialLocation(self):
        print("PYTHON TEST: spatialLocation")
        image = "./TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_spatloc_eucdist.kea"
        eastings = "Easts"
        northings = "Norths"
        rastergis.spatialLocation(image, eastings, northings)

    def testEucDistFromFeat(self):
        print("PYTHON TEST: eucDistFromFeature")
        image = "./TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_spatloc_eucdist.kea"
        feature=10
        outfield="DistTo10"
        fields = ["Easts", "Norths"]
        rastergis.eucDistFromFeature(image, feature, outfield, fields)

    def testFindTopN(self):
        print("PYTHON TEST: findTopN")
        image = "./TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_spatloc_eucdist.kea"
        spatialdist="DistTo10"
        metricdist="RedAvg"
        n=10
        distthreshold=100
        outfield="Top10Feats"
        rastergis.findTopN(image, spatialdist, metricdist, outfield, n, distthreshold)

    def testFindSpecClose(self):
        print("PYTHON TEST: findSpecClose")
        image = "./TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_spatloc_eucdist.kea"
        spatialdist="DistTo10"
        metricdist="RedAvg"
        n=10
        specdistthreshold=20
        spatdistthreshold=100
        outfield="Closest2Feat10"
        rastergis.findSpecClose(image, spatialdist, metricdist, outfield, specdistthreshold, spatdistthreshold)
    """ 
    def testPopulateRATWithStats(self):
        print("PYTHON TEST: populateRATWithStats")
        clumps="./TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_popstats.kea"
        input="./Rasters/injune_p142_casi_sub_utm.kea"
        bs = []
        bs.append(rastergis.BandAttStats(band=1, minField="b1Min", maxField="b1Max", meanField="b1Mean", sumField="b1Sum", stdDevField="b1StdDev"))
        bs.append(rastergis.BandAttStats(band=2, minField="b2Min", maxField="b2Max", meanField="b2Mean", sumField="b2Sum", stdDevField="b2StdDev"))
        bs.append(rastergis.BandAttStats(band=3, minField="b3Min", maxField="b3Max", meanField="b3Mean", sumField="b3Sum", stdDevField="b3StdDev"))
        rastergis.populateRATWithStats(input, clumps, bs)

    def testPopulateRATWithPercentiles(self):
        print("PYTHON TEST: populateRATWithPercentiles")
        clumps = "./TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_popstats.kea"
        input = "./Rasters/injune_p142_casi_sub_utm.kea"
        bp = []
        bp.append(rastergis.BandAttPercentiles(percentile=25.0, fieldName="B1Per25"))
        bp.append(rastergis.BandAttPercentiles(percentile=50.0, fieldName="B1Per50"))
        bp.append(rastergis.BandAttPercentiles(percentile=75.0, fieldName="B1Per75"))
        rastergis.populateRATWithPercentiles(input, clumps, 1, bp)

    def testExport2Ascii(self):
        print("PYTHON TEST: export2Ascii")
        table="./RATS/injune_p142_casi_sub_utm_segs.kea"
        output="./TestOutputs/RasterGIS/injune_p142_casi_rgb_exportascii.txt"
        fields = ["b1Mean", "b2Mean", "b3Mean"]
        rastergis.export2Ascii(table, output, fields)

    def testExportCol2GDALImage(self):
        print("PYTHON TEST: exportCol2GDALImage")
        clumps="./RATS/injune_p142_casi_sub_utm_segs.kea"
        output="./TestOutputs/RasterGIS/injune_p142_casi_b1Mean_export.kea"
        gdalformat = "KEA"
        dataType = rsgislib.TYPE_32FLOAT
        field = "b1Mean"
        rastergis.exportCol2GDALImage(clumps, output, gdalformat, dataType, field)

    def testExportCols2GDALImage(self):
        print("PYTHON TEST: exportCols2GDALImage")
        clumps="./RATS/injune_p142_casi_sub_utm_segs.kea"
        output="./TestOutputs/RasterGIS/injune_p142_casi_b1-2Mean_export.kea"
        gdalformat = "KEA"
        dataType = rsgislib.TYPE_32FLOAT
        fields = ['b1Mean','b2Mean']
        rastergis.exportCols2GDALImage(clumps, output, gdalformat, dataType, fields)

    def testExportColourClasses(self):
        print('PYTHON TEST: colourClasses')
        clumps='./TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_col.kea'
        field = 'outClass'
        classcolours = {}
        colourCat = collections.namedtuple('ColourCat', ['red', 'green', 'blue', 'alpha'])
        classcolours[0] = colourCat(red=200, green=50, blue=50, alpha=255)
        classcolours[1] = colourCat(red=200, green=240, blue=50, alpha=255)
        rastergis.colourClasses(clumps, field, classcolours)
        
    def testExportColourClassesStr(self):
        print('PYTHON TEST: colourClasses (String)')
        clumps='./TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_col_str.kea'
        field = 'outClassStr'
        classcolours = {}
        colourCat = collections.namedtuple('ColourCat', ['red', 'green', 'blue', 'alpha'])
        classcolours['ClassA'] = colourCat(red=200, green=50, blue=50, alpha=255)
        classcolours['ClassB'] = colourCat(red=200, green=240, blue=50, alpha=255)
        rastergis.colourClasses(clumps, field, classcolours)

    """
    def testFindNeighbours(self):
        print('PYTHON TEST: findNeighbours')
        input = './TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_neighbours.kea'
        rastergis.findNeighbours(input)

    def testFindBoundaryPixels(self):
        print('PYTHON TEST: findBoundaryPixels')
        clumps='./RATS/injune_p142_casi_sub_utm_segs.kea'
        output='./TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_borpxls.kea'
        gdalformat='KEA'
        rastergis.findBoundaryPixels(clumps, output, gdalformat)

    def testCalcBorderLength(self):
        print('PYTHON TEST: calcBorderLength')
        clumps='./TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_borlen.kea'
        ignorezeroedges=True
        colname='borderLen'
        rastergis.calcBorderLength(clumps, ignorezeroedges, colname)

    def testCalcShapeIndices(self):
        print('PYTHON TEST: calcShapeIndices')
        clumps = './TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_shape.kea'
        shapes = []
        shapes.append(rastergis.ShapeIndex(colName='Area', idx=rsgislib.SHAPE_SHAPEAREA))
        shapes.append(rastergis.ShapeIndex(colName='Length', idx=rsgislib.SHAPE_LENGTH))
        shapes.append(rastergis.ShapeIndex(colName='Width', idx=rsgislib.SHAPE_WIDTH))
        shapes.append(rastergis.ShapeIndex(colName='BorderLengthIdx', idx=rsgislib.SHAPE_BORDERLENGTH))
        shapes.append(rastergis.ShapeIndex(colName='Asymmetry', idx=rsgislib.SHAPE_ASYMMETRY))
        shapes.append(rastergis.ShapeIndex(colName='ShapeIndex', idx=rsgislib.SHAPE_SHAPEINDEX))
        shapes.append(rastergis.ShapeIndex(colName='Density', idx=rsgislib.SHAPE_DENSITY))
        shapes.append(rastergis.ShapeIndex(colName='LengthWidthRatio', idx=rsgislib.SHAPE_LENGTHWIDTH))
        shapes.append(rastergis.ShapeIndex(colName='BorderIndex', idx=rsgislib.SHAPE_BORDERINDEX))
        shapes.append(rastergis.ShapeIndex(colName='Compactness', idx=rsgislib.SHAPE_COMPACTNESS))
        shapes.append(rastergis.ShapeIndex(colName='MainDirection', idx=rsgislib.SHAPE_MAINDIRECTION))
        rastergis.calcShapeIndices(clumps, shapes)
    """

    def testFindChangeClumpsFromStdDev(self):
        print("PYTHON TEST: findChangeClumpsFromStdDev")
        clumps="./TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_change.kea"
        ChangeFeat = collections.namedtuple('ChangeFeats', ['name', 'outName', 'threshold'])
        changeFeatVals = []
        attributes = ['b1Mean','b2Mean']
        changeFeatVals.append(ChangeFeat(name="ClassA", outName=1, threshold=1.0))
        changeFeatVals.append(ChangeFeat(name="ClassB", outName=2, threshold=1))
        rastergis.findChangeClumpsFromStdDev(clumps, "outClassStr", "Change", attributes, changeFeatVals)
    
    def testPopulateStats(self):
        print("PYTHON TEST: populateStats")
        clumps="./TestOutputs/RasterGIS/injune_p142_casi_sub_utm_segs_nostats_addstats.kea"
        pyramids=True
        colourtable=True
        rastergis.populateStats(clumps, colourtable, pyramids)

    # Image Utils 
    
    def testCreateTiles(self):
        print("PYTHON TEST: createTiles")
        inputImage = path + "Rasters/injune_p142_casi_sub_utm.kea"
        outBase = './TestOutputs/Tiles/injune_p142_casi_sub_utm'
        width = 100
        height = width
        overlap = 5
        offsettiling = 0
        gdalformat = "KEA"
        dataType = rsgislib.TYPE_32INT
        ext='kea'
        imageutils.createTiles(inputImage, outBase, width, height, overlap, offsettiling, gdalformat, dataType, ext)

    def testCreateImageMosaic(self):
        print("PYTHON TEST: createImageMosaic")
        inputList = ['./TestOutputs/Tiles/injune_p142_casi_sub_utm_tile0.kea','./TestOutputs/Tiles/injune_p142_casi_sub_utm_tile1.kea','./TestOutputs/Tiles/injune_p142_casi_sub_utm_tile2.kea','./TestOutputs/Tiles/injune_p142_casi_sub_utm_tile3.kea']
        outImage = './TestOutputs/injune_p142_casi_sub_utm_remosaic.kea'
        backgroundVal = 0.
        skipVal = 0.
        skipBand = 1
        overlapBehaviour = 0
        gdalformat = "KEA"
        dataType = rsgislib.TYPE_32FLOAT
        imageutils.createImageMosaic(inputList, outImage, backgroundVal, skipVal, skipBand, overlapBehaviour, gdalformat, dataType)

    def testIncludeImages(self):
        print("PYTHON TEST: includeImages")
        baseImage = './TestOutputs/injune_p142_casi_sub_utm_remosaic.kea'
        inputList = ['./TestOutputs/Tiles/injune_p142_casi_sub_utm_tile0.kea','./TestOutputs/Tiles/injune_p142_casi_sub_utm_tile1.kea']
        imageutils.includeImages(baseImage, inputList)
    
    def testPopImageStats(self):
        print("PYTHON TEST: popImageStats")
        inputImage = './TestOutputs/injune_p142_casi_sub_utm.kea'
        imageutils.popImageStats(inputImage,True,0.,True)
        
    def testSubset(self):
        print("PYTHON TEST: subset")
        inputImage = './Rasters/injune_p142_casi_sub_utm.kea'
        inputVector = './Vectors/injune_p142_plot_location_utm.shp'
        outputImage = './TestOutputs/injune_p142_casi_sub_utm_subset.kea'
        gdalformat = "KEA"
        dataType = rsgislib.TYPE_32FLOAT
        imageutils.subset(inputImage, inputVector, outputImage, gdalformat, dataType)

    def testSubset2Polys(self):
        print("PYTHON TEST: subset2polys")
        inputImage = './Rasters/injune_p142_casi_sub_utm.kea'
        inputVector = './Vectors/injune_p142_plot_location_utm.shp'
        attribute = 'PLOTNO'
        outputImageBase = './TestOutputs/injune_p142_casi_sub_utm_subset_polys_'
        gdalformat = 'KEA'
        dataType = rsgislib.TYPE_32FLOAT
        ext = 'kea'
        outFiles = imageutils.subset2polys(inputImage, inputVector, attribute, outputImageBase, gdalformat, dataType, ext)

    def testSubset2Img(self):
        print("PYTHON TEST: subset2img")
        inputImage = './Rasters/injune_p142_casi_sub_utm.kea'
        inputROIImage = './TestOutputs/injune_p142_casi_sub_utm_subset.kea'
        outputImage = './TestOutputs/injune_p142_casi_sub_utm_subset2img.kea'
        gdalformat = "KEA"
        dataType = rsgislib.TYPE_32FLOAT
        imageutils.subset2img(inputImage, inputROIImage, outputImage, gdalformat, dataType)

    def testStackBands(self):
        print("PYTHON TEST: stackImageBands")
        imageList = ['./Rasters/injune_p142_casi_sub_utm_single_band.vrt','./Rasters/injune_p142_casi_sub_utm_single_band.vrt']
        bandNamesList = ['Image1','Image2']
        outputImage = './TestOutputs/injune_p142_casi_sub_stack.kea'
        gdalformat = "KEA"
        dataType = rsgislib.TYPE_32FLOAT
        imageutils.stackImageBands(imageList, bandNamesList, outputImage, None, 0, gdalformat, dataType)

    def testCreateCopyImage(self):
        print("PYTHON TEST: createCopyImage")
        inputImage = './Rasters/injune_p142_casi_sub_utm.kea'
        outputImage = './TestOutputs/injune_p142_casi_sub_utm_blank.kea'
        gdalformat = 'KEA'
        dataType = rsgislib.TYPE_32FLOAT
        imageutils.createCopyImage(inputImage, outputImage, 1, 3, gdalformat, dataType)

    def testStackStats(self):
        print("PYTHON TEST: stackStats")
        inputImage = './Rasters/injune_p142_casi_sub_utm.kea'
        outputImage = './TestOutputs/injune_p142_casi_sub_utm_stackStats.kea'
        gdalformat = 'KEA'
        dataType = rsgislib.TYPE_32FLOAT
        imageutils.stackStats(inputImage, outputImage, None, 'mean', gdalformat, dataType)

    def testStretchImage(self):
        print("PYTHON TEST: stretchImage")
        inputImage = './Rasters/injune_p142_casi_sub_utm.kea'
        outputImage = './TestOutputs/injune_p142_casi_sub_utm_2sd.kea'
        gdalformat = 'KEA'
        dataType = rsgislib.TYPE_8INT
        imageutils.stretchImage(inputImage, outputImage, False, "", True, False, gdalformat, dataType, imageutils.STRETCH_LINEARSTDDEV, 2)

    def testSetBandNames(self):
        print("PYTHON TEST: setBandNames")
        inputImage = './TestOutputs/injune_p142_casi_sub_utm.kea'
        bandNames = ['446nm','530nm','549nm','569nm','598nm','633nm','680nm','696nm','714nm','732nm','741nm','752nm','800nm','838nm']
        imageutils.setBandNames(inputImage, bandNames)

    def testGetRSGISLibDataType(self):
        print("PYTHON TEST: getRSGISLibDataType")
        inputImage = './Rasters/injune_p142_casi_sub_utm.kea'
        imageutils.getRSGISLibDataType(inputImage)

    def testGetGDALDataType(self):
        print("PYTHON TEST: getGDALDataType")
        inputImage = './Rasters/injune_p142_casi_sub_utm.kea'
        imageutils.getGDALDataType(inputImage)

     # Zonal Stats

    def testPointValue2SHP(self):
        # Not all pixels are within image - should print warnings but pass test.
        print("PYTHON TEST: pointValues2SHP")
        inputImage = './Rasters/injune_p142_casi_sub_utm.kea'
        inputVector = './Vectors/injune_p142_stem_locations.shp'
        outputVector = './TestOutputs/injune_p142_stem_locations_stats.shp'
        zonalstats.pointValue2SHP(inputImage, inputVector, outputVector, True, True)

    def testPointValue2TXT(self):
        # Not all pixels are within image - should print warnings but pass test.
        print("PYTHON TEST: pointValues2TXT")
        inputImage = './Rasters/injune_p142_casi_sub_utm.kea'
        inputVector = './Vectors/injune_p142_stem_locations.shp'
        outputTxt = './TestOutputs/injune_p142_stem_locations_stats_txt.csv'
        zonalstats.pointValue2TXT(inputImage, inputVector, outputTxt, True)
        
    def testPixelStats2SHP(self, allCases=False):
        """ Check for three cases, all polygons inside image (1) and polygons outside image (2) """
        print("PYTHON TEST: pixelStats2SHP")
        inputImage1 = './Rasters/injune_p142_casi_sub_utm.kea'
        inputImage2 = './Rasters/injune_p142_casi_sub_left_utm.kea'
        inputImage3 = './Rasters/injune_p142_casi_sub_right_utm.kea'
        inputVector = './Vectors/injune_p142_crowns_utm.shp'
        outputVector1 = './TestOutputs/injune_p142_casi_sub_utm_stats1.shp'
        outputVector2 = './TestOutputs/injune_p142_casi_sub_utm_stats2.shp'
        outputVector3 = './TestOutputs/injune_p142_casi_sub_utm_stats3.shp'
        zonalattributes = zonalstats.ZonalAttributes(minThreshold=0, maxThreshold=10000, calcCount=True, calcMin=True, calcMax=True, calcMean=True, calcStdDev=True, calcMode=False, calcSum=True)
        zonalstats.pixelStats2SHP(inputImage1, inputVector, outputVector1, zonalattributes, True, True, True, zonalstats.METHOD_POLYCONTAINSPIXELCENTER)
        if allCases:
            zonalstats.pixelStats2SHP(inputImage2, inputVector, outputVector2, zonalattributes, True, True, True, zonalstats.METHOD_POLYCONTAINSPIXELCENTER)
            zonalstats.pixelStats2SHP(inputImage3, inputVector, outputVector3, zonalattributes, True, True, True, zonalstats.METHOD_POLYCONTAINSPIXELCENTER)
    
    def testPixelStats2TXT(self):
        print("PYTHON TEST: pixelStats2TXT")
        inputImage = './Rasters/injune_p142_casi_sub_left_utm.kea'
        inputVector = './Vectors/injune_p142_crowns_utm.shp'
        outputTxt = './TestOutputs/injune_p142_casi_sub_utm_stats_txt.csv'
        zonalattributes = zonalstats.ZonalAttributes(minThreshold=0, maxThreshold=10000, calcCount=True, calcMin=True, calcMax=True, calcMean=True, calcStdDev=True, calcMode=False, calcSum=True)
        zonalstats.pixelStats2TXT(inputImage, inputVector, outputTxt, zonalattributes, True, True, zonalstats.METHOD_POLYCONTAINSPIXELCENTER)
    
    def testPixelVals2TXT(self):
        print("PYTHON TEST: pixelVals2TXT")
        inputImage = './Rasters/injune_p142_casi_sub_utm.kea'
        inputVector = './Vectors/injune_p142_crowns_utm.shp'
        outputTxtBase = './TestOutputs/ZonalTXT/injune_p142_casi_sub_utm_txt'
        zonalstats.pixelVals2TXT(inputImage, inputVector, outputTxtBase, 'FID', True, zonalstats.METHOD_POLYCONTAINSPIXELCENTER)    
    
    def testImageZone2HDF(self):
        print("PYTHON TEST: pixelVals2TXT")
        inputimage = './Rasters/injune_p142_casi_sub_utm.kea'
        inputvector = './Vectors/injune_p142_crowns_withincasi_utm.shp'
        outputHDF = './TestOutputs/InjuneP142.hdf'
        zonalstats.imageZoneToHDF(inputimage, inputvector, outputHDF, True, zonalstats.METHOD_POLYCONTAINSPIXELCENTER)
        
    # Image Registration
    def testBasicRegistration(self):
        print("PYTHON TEST: basicregistration")
        reference = './Rasters/injune_p142_casi_sub_utm_single_band.vrt'        
        floating = './Rasters/injune_p142_casi_sub_utm_single_band_offset3x3y.vrt'
        pixelGap = 50 
        threshold = 0.4
        window = 100
        search = 5
        stddevRef = 2
        stddevFloat = 2
        subpixelresolution = 4
        metric = imageregistration.METRIC_CORELATION
        outputType = imageregistration.TYPE_RSGIS_IMG2MAP
        output = './TestOutputs/injune_p142_casi_sub_utm_tie_points_basic.txt'
        imageregistration.basicregistration(reference, floating, pixelGap, threshold, window, search, stddevRef, stddevFloat, subpixelresolution, metric, outputType, output)
        
    def testSingleLayerRegistration(self):
        print("PYTHON TEST: singlelayerregistration")
        reference = './Rasters/injune_p142_casi_sub_utm_single_band.vrt'        
        floating = './Rasters/injune_p142_casi_sub_utm_single_band_offset3x3y.vrt'
        pixelGap = 50 
        threshold = 0.4
        window = 100
        search = 5
        stddevRef = 2
        stddevFloat = 2
        subpixelresolution = 4
        distanceThreshold = 100
        maxiterations = 10
        movementThreshold = 0.01 
        pSmoothness = 2    
        metric = imageregistration.METRIC_CORELATION
        outputType = imageregistration.TYPE_RSGIS_IMG2MAP
        output = './TestOutputs/injune_p142_casi_sub_utm_tie_points_singlelayer.txt'
        imageregistration.singlelayerregistration(reference, floating, pixelGap, threshold, window, search, stddevRef, stddevFloat, subpixelresolution, distanceThreshold, maxiterations, movementThreshold, pSmoothness, metric, outputType, output)
        
    def testTriangularWarp(self):        
        inputImage = './Rasters/injune_p142_casi_sub_utm_single_band_offset3x3y.vrt'
        inputGCPs = './TestOutputs/injune_p142_casi_sub_utm_tie_points_basic.txt'
        outputImage = './TestOutputs/injune_p142_casi_sub_utm_single_band_offset3x3y_twarp.kea'
        wktStringFile = './Vectors/injune_p142_crowns_utm.prj'
        resolution = 1
        gdalformat = 'KEA'
        imageregistration.triangularwarp(inputImage,inputGCPs, outputImage, wktStringFile, resolution, gdalformat)
        
    def testNNWarp(self):        
        inputImage = './Rasters/injune_p142_casi_sub_utm_single_band_offset3x3y.vrt'
        inputGCPs = './TestOutputs/injune_p142_casi_sub_utm_tie_points_basic.txt'
        outputImage = './TestOutputs/injune_p142_casi_sub_utm_single_band_offset3x3y_nnwarp.kea'
        wktStringFile = './Vectors/injune_p142_crowns_utm.prj'
        resolution = 1
        gdalformat = 'KEA'
        imageregistration.nnwarp(inputImage,inputGCPs, outputImage, wktStringFile, resolution, gdalformat)

    def testPolyWarp(self):        
        inputImage = './Rasters/injune_p142_casi_sub_utm_single_band_offset3x3y.vrt'
        inputGCPs = './TestOutputs/injune_p142_casi_sub_utm_tie_points_basic.txt'
        outputImage = './TestOutputs/injune_p142_casi_sub_utm_single_band_offset3x3y_polywarp.kea'
        wktStringFile = './Vectors/injune_p142_crowns_utm.prj'
        resolution = 1
        polyOrder = 1
        gdalformat = 'KEA'
        imageregistration.polywarp(inputImage,inputGCPs, outputImage, wktStringFile, resolution, polyOrder, gdalformat)

    def testGCP2GDAL(self):
        print("PYTHON TEST: gcps2gdal")
        inputImage = './Rasters/injune_p142_casi_sub_utm_single_band_offset3x3y.vrt'
        inputGCPs = './TestOutputs/injune_p142_casi_sub_utm_tie_points_basic.txt'
        outputImage = './TestOutputs/injune_p142_casi_sub_utm_single_band_offset3x3y_gcps.kea'
        gdalformat = "KEA"
        dataType = rsgislib.TYPE_32INT
        imageregistration.gcp2gdal(inputImage,inputGCPs, outputImage, gdalformat, dataType)
    
    # Vector utils
    def testRemoveAttributes(self):
        print("PYTHON TEST: removeattributes")
        inputVector = './Vectors/injune_p142_stem_locations.shp'
        outputVector = './TestOutputs/injune_p142_stem_locations_noatts.shp'
        vectorutils.removeattributes(inputVector, outputVector, True)

    def testBufferVector(self):
        print("PYTHON TEST: buffervector")
        inputVector = './Vectors/injune_p142_stem_locations.shp'
        lyrName = 'injune_p142_stem_locations'
        outputVector = './TestOutputs/injune_p142_stem_locations_1mbuffer.shp'
        bufferDist = 1
        vectorutils.buffervector(inputVector, lyrName, outputVector, 'ESRI Shapefile', bufferDist)

    def testPrintPolyGeom(self):
        print("PYTHON TEST: buffervector")
        inputVector = './Vectors/injune_p142_psu_utm.shp'
        vectorutils.printpolygeom(inputVector)

    def testFindReplaceText(self):
        print("PYTHON TEST: findreplacetext")
        inputVector = './TestOutputs/injune_p142_psu_utm.shp'
        attribute = 'PSU'
        find = '142'
        replace = '142'
        vectorutils.findreplacetext(inputVector, attribute, find, replace)

    def testCalcArea(self):
        print("PYTHON TEST: calcarea")
        inputVector = './Vectors/injune_p142_psu_utm.shp'
        outputVector = './TestOutputs/injune_p142_psu_utm_area.shp'
        vectorutils.calcarea(inputVector, outputVector, True)

    def testPolygonsInPolygon(self):
        print("PYTHON TEST: polygonsInPolygon")
        inputVector = './Vectors/injune_p142_stem_locations.shp'
        coverVector = './Vectors/injune_p142_psu_utm.shp'
        outDIR = './TestOutputs'
        attribute = 'PSU'
        vectorutils.polygonsInPolygon(inputVector, coverVector, outDIR, attribute, True)


    # Image filter
    def testFilter(self, allFilters=False):
        print("PYTHON TEST: filter")
        inputImage = './Rasters/injune_p142_casi_sub_utm_single_band.vrt'
        outputImageBase = './TestOutputs/injune_p142_casi_sub_utm_single_band'
        gdalFormat = 'KEA'
        outExt = 'kea'
        dataType = rsgislib.TYPE_32FLOAT
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'GaussianSmooth', fileEnding = 'gausmooth', size=3, stddevX = 1., stddevY = 1, angle = 0.) )
        if allFilters:
            filters.append(imagefilter.FilterParameters(filterType = 'Gaussian1st', fileEnding = 'gau1st', size=3, stddevX = 1., stddevY = 1, angle = 0.) )
            filters.append(imagefilter.FilterParameters(filterType = 'Gaussian2nd', fileEnding = 'gau1st', size=3, stddevX = 1., stddevY = 1, angle = 0.) )
            filters.append(imagefilter.FilterParameters(filterType = 'Laplacian', fileEnding = 'laplacian', size=3, stddev = 1) )
            filters.append(imagefilter.FilterParameters(filterType = 'Sobel', fileEnding = 'sobelx', option = 'x') )
            filters.append(imagefilter.FilterParameters(filterType = 'Sobel', fileEnding = 'sobely', option = 'y') )
            filters.append(imagefilter.FilterParameters(filterType = 'Sobel', fileEnding = 'sobelxy', option = 'xy') )
            filters.append(imagefilter.FilterParameters(filterType = 'Prewitt', fileEnding = 'prewittx', option = 'x') )
            filters.append(imagefilter.FilterParameters(filterType = 'Prewitt', fileEnding = 'prewitty', option = 'y') )
            filters.append(imagefilter.FilterParameters(filterType = 'Prewitt', fileEnding = 'prewittxy', option = 'xy') )
            filters.append(imagefilter.FilterParameters(filterType = 'Mean', fileEnding = 'mean', size=3) )
            filters.append(imagefilter.FilterParameters(filterType = 'Median', fileEnding = 'median', size=3) )
            filters.append(imagefilter.FilterParameters(filterType = 'Mode', fileEnding = 'mode', size=3) )
            filters.append(imagefilter.FilterParameters(filterType = 'StdDev', fileEnding = 'stddev', size=3) )
            filters.append(imagefilter.FilterParameters(filterType = 'Range', fileEnding = 'range', size=3) )
            filters.append(imagefilter.FilterParameters(filterType = 'CoeffOfVar', fileEnding = 'coeffofvar', size=3) )
            filters.append(imagefilter.FilterParameters(filterType = 'Min', fileEnding = 'min', size=3) )
            filters.append(imagefilter.FilterParameters(filterType = 'Max', fileEnding = 'max', size=3) )
            filters.append(imagefilter.FilterParameters(filterType = 'Total', fileEnding = 'total', size=3) )
            filters.append(imagefilter.FilterParameters(filterType = 'Kuwahara', fileEnding = 'kuwahara', size=3) )
            filters.append(imagefilter.FilterParameters(filterType = 'Lee', fileEnding = 'lee', size=3, nLooks=3) )
            filters.append(imagefilter.FilterParameters(filterType = 'NormVar', fileEnding = 'normvar', size=3) )
            filters.append(imagefilter.FilterParameters(filterType = 'NormVarSqrt', fileEnding = 'normvarsqrt', size=3) )
            filters.append(imagefilter.FilterParameters(filterType = 'NormVarLn', fileEnding = 'normvarln', size=3) )
            filters.append(imagefilter.FilterParameters(filterType = 'TextureVar', fileEnding = 'texturevar', size=3) )

        imagefilter.applyfilters(inputImage, outputImageBase, filters, gdalFormat, outExt, dataType)
    
    def testLeungMalikFilterBank(self):
        inputImage = './Rasters/injune_p142_casi_sub_utm_single_band.vrt'
        outputImageBase = './TestOutputs/injune_p142_casi_sub_utm_single_band'
        gdalFormat = 'KEA'
        outExt = 'kea'
        dataType = rsgislib.TYPE_32FLOAT
        imagefilter.LeungMalikFilterBank(inputImage, outputImageBase, gdalFormat, outExt, dataType)

    # Segmentation
    def testUnionOfClumps(self):
        clumps1='./RATS/injune_p142_casi_sub_utm_segs.kea'
        clumps2='./RATS/injune_p142_casi_sub_utm_segs.kea'
        outputimage = './TestOutputs/injune_p142_casi_sub_utm_segs_union_test.kea'
        inputimagepaths = [clumps1,clumps2]
        gdalFormat = 'KEA'
        
        segmentation.unionOfClumps(outputimage, gdalFormat, inputimagepaths, 0, False)

    def testRunShepherdSegmentation(self):
        inputImage = './Rasters/injune_p142_casi_sub_utm.kea'
        clumpsFile = './TestOutputs/injune_p142_casi_sub_utm_seg_test.kea'
        meanImage = './TestOutputs/injune_p142_casi_sub_utm_seg_test_mean.kea'

        segmentation.segutils.runShepherdSegmentation(inputImage, clumpsFile,
                       meanImage, numClusters=100, minPxls=100)

    # Tools
    def testMetres2Degrees(self):
        print(tools.metres_to_degrees(52,1,1))

    def testDegrees2Metres(self):
        print(tools.degrees_to_metres(52,1,1))

if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("--all", action='store_true', default=False, help="Run all tests")
    parser.add_argument("--imagecalc", action='store_true', default=False, help="Run imagecalc tests")
    parser.add_argument("--imagefilter", action='store_true', default=False, help="Run imagefilter tests")
    parser.add_argument("--imageregistration", action='store_true', default=False, help="Run imageregistration tests")
    parser.add_argument("--imageutils", action='store_true', default=False, help="Run imageutils tests")
    parser.add_argument("--rastergis", action='store_true', default=False, help="Run rastergis tests")
    parser.add_argument("--segmentation", action='store_true', default=False, help="Run segmentation tests")
    parser.add_argument("--tools", action='store_true', default=False, help="Run tools tests")
    parser.add_argument("--vectorutils", action='store_true', default=False, help="Run vectorutils tests")
    parser.add_argument("--zonalstats", action='store_true', default=False, help="Run zonalstats tests")
    args = parser.parse_args() 

    t = RSGISTests()

    """ Check directory structure and copy Data """
    t.checkDIRStructure()
    t.copyData()
    
    if args.all or args.imagecalc:
        """ ImageCalc functions """
        t.tryFuncAndCatch(t.testNormalise1)
        t.tryFuncAndCatch(t.testNormalise2)
        t.tryFuncAndCatch(t.testCorrelation)
        t.tryFuncAndCatch(t.testCovariance1)
        #t.tryFuncAndCatch(t.testCovariance2) # Missing matrix files.
        t.tryFuncAndCatch(t.testCalcRMSE)
        t.tryFuncAndCatch(t.testMeanVector)
        t.tryFuncAndCatch(t.testPCA)
        t.tryFuncAndCatch(t.testStandardise)
        t.tryFuncAndCatch(t.testBandMath)
        t.tryFuncAndCatch(t.testImageMaths)
        t.tryFuncAndCatch(t.testReplaceValuesLessThan)
        t.tryFuncAndCatch(t.testUnitArea)
        t.tryFuncAndCatch(t.testMovementSpeed)
        t.tryFuncAndCatch(t.testCountValsInCol)
        t.tryFuncAndCatch(t.testDist2Geoms)
        t.tryFuncAndCatch(t.testImgDist2Geoms)
        t.tryFuncAndCatch(t.testImgCalcDist)
        t.tryFuncAndCatch(t.testImageBandStats)
        t.tryFuncAndCatch(t.testImageBandStatsIgnoreZeros)
        t.tryFuncAndCatch(t.testImageStats)
        t.tryFuncAndCatch(t.testImageStatsIgnoreZeros)
        t.tryFuncAndCatch(t.testUnconLinearSpecUnmix)
        t.tryFuncAndCatch(t.testExhConLinearSpecUnmix)
        t.tryFuncAndCatch(t.testConSum1LinearSpecUnmix)
        t.tryFuncAndCatch(t.testNnConSum1LinearSpecUnmix)
        t.tryFuncAndCatch(t.testKMeansCentres)
        t.tryFuncAndCatch(t.testIsoDataClustering)
        t.tryFuncAndCatch(t.testAllBandsEqualTo)
        t.tryFuncAndCatch(t.testHistogram)
        t.tryFuncAndCatch(t.testBandPercentile)
        t.tryFuncAndCatch(t.testMahalanobisDistWindow)
        t.tryFuncAndCatch(t.testMahalanobisDistImg2Window)
        t.tryFuncAndCatch(t.testCalcPxlColStats)
        t.tryFuncAndCatch(t.testCorrelationWindow)
        t.tryFuncAndCatch(t.testImagePixelLinearFit)
        
    if args.all or args.imageutils:
        
        """ ImageUtils functions """
        t.tryFuncAndCatch(t.testCreateTiles)
        t.tryFuncAndCatch(t.testCreateImageMosaic)
        t.tryFuncAndCatch(t.testIncludeImages)
        t.tryFuncAndCatch(t.testPopImageStats)
        t.tryFuncAndCatch(t.testSubset)
        t.tryFuncAndCatch(t.testSubset2Polys)
        t.tryFuncAndCatch(t.testSubset2Img)
        t.tryFuncAndCatch(t.testStackBands)
        t.tryFuncAndCatch(t.testStackStats)
        t.tryFuncAndCatch(t.testCreateCopyImage)
        t.tryFuncAndCatch(t.testStretchImage)
        t.tryFuncAndCatch(t.testSetBandNames)
        t.tryFuncAndCatch(t.testGetRSGISLibDataType)
        t.tryFuncAndCatch(t.testGetGDALDataType)
        
    if args.all or args.rastergis:
    
        """ RasterGIS functions """
        t.tryFuncAndCatch(t.testCopyGDLATT)
        t.tryFuncAndCatch(t.testCopyGDLATTColumns)
        t.tryFuncAndCatch(t.testExportColourClasses)
        t.tryFuncAndCatch(t.testExportColourClassesStr)
        #t.tryFuncAndCatch(t.testSpatialLocation)
        #t.tryFuncAndCatch(t.testEucDistFromFeat)
        #t.tryFuncAndCatch(t.testFindTopN)
        #t.tryFuncAndCatch(t.testFindSpecClose)
        t.tryFuncAndCatch(t.testPopulateRATWithStats)
        t.tryFuncAndCatch(t.testPopulateRATWithPercentiles)
        t.tryFuncAndCatch(t.testExport2Ascii)
        t.tryFuncAndCatch(t.testExportCol2GDALImage)
        t.tryFuncAndCatch(t.testExportCols2GDALImage)
        t.tryFuncAndCatch(t.testPopulateStats)
        #t.tryFuncAndCatch(t.testFindNeighbours)
        #t.tryFuncAndCatch(t.testFindBoundaryPixels)
        #t.tryFuncAndCatch(t.testCalcBorderLength)
        #t.tryFuncAndCatch(t.testCalcShapeIndices)
        t.tryFuncAndCatch(t.testFindChangeClumpsFromStdDev)
        t.tryFuncAndCatch(t.testCopyGDLATTColumns)
        
    if args.all or args.zonalstats:
        
        """ Zonal Stats functions """
        t.tryFuncAndCatch(t.testPointValue2SHP)
        t.tryFuncAndCatch(t.testPointValue2TXT)
        t.tryFuncAndCatch(t.testPixelStats2SHP)
        t.tryFuncAndCatch(t.testPixelStats2TXT)
        t.tryFuncAndCatch(t.testPixelVals2TXT)
        t.tryFuncAndCatch(t.testImageZone2HDF)
        
    if args.all or args.imageregistration:
        
        """ Image Registration functions """
        t.tryFuncAndCatch(t.testBasicRegistration)
        t.tryFuncAndCatch(t.testSingleLayerRegistration)
        t.tryFuncAndCatch(t.testGCP2GDAL)
        t.tryFuncAndCatch(t.testTriangularWarp)
        t.tryFuncAndCatch(t.testNNWarp)
        t.tryFuncAndCatch(t.testPolyWarp)
    
    if args.all or args.vectorutils:
        
        """ Vector Utils functions """
        t.tryFuncAndCatch(t.testRemoveAttributes)
        t.tryFuncAndCatch(t.testBufferVector)
        t.tryFuncAndCatch(t.testPrintPolyGeom)
        t.tryFuncAndCatch(t.testFindReplaceText)
        t.tryFuncAndCatch(t.testCalcArea)
        t.tryFuncAndCatch(t.testPolygonsInPolygon)

    if args.all or args.imagefilter:
        """ Image filter functions """ 
        t.tryFuncAndCatch(t.testFilter)
        #t.tryFuncAndCatch(t.testLeungMalikFilterBank) # Skip as it takes a while
    
    if args.all or args.segmentation:
        """ Image filter functions """ 
        t.tryFuncAndCatch(t.testUnionOfClumps)
        t.tryFuncAndCatch(t.testRunShepherdSegmentation)


    if args.all or args.tools:
        t.tryFuncAndCatch(t.testMetres2Degrees)    
        t.tryFuncAndCatch(t.testDegrees2Metres)    

    t.removeTestFiles()

    print("%s TESTS COMPLETED - %s FAILURES LISTED BELOW:"%(t.numTests, len(t.failures)))
    returnCode = 0 # success state
    if(len(t.failures)):
        returnCode = 1 # error state so calling script can easily check
        for failure in t.failures:
            print("FAIL: %s"%failure)

    # return whether all tests succeeded or not
    sys.exit(returnCode)

    

    







#!/usr/bin/env python

#############################################
# rgsisvectools.py
#
#  Copyright 2018 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
# Purpose:  Get information on vector files/layers. 
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 17/12/2018
# Version: 1.0
# 
#############################################

import rsgislib.vectorutils

import argparse

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--vecfile", type=str, required=True, help="Input vector file.")
    parser.add_argument("--veclyr", type=str, required=False, help="Input vector file layer.")
    parser.add_argument("--lyrs", action='store_true', default=False, help="List the layers in vector file.")
    parser.add_argument("--cols", action='store_true', default=False, help="List the columns in vector layer.")
    args = parser.parse_args()
    
    if args.lyrs:
        veclyrs = rsgislib.vectorutils.getVecLyrsLst(args.vecfile)
        i = 1
        for lyr in veclyrs:
            print("{0}: {1}".format(i, lyr))
            i = i + 1
    if args.cols:
        veccols = rsgislib.vectorutils.lstveclyrcols(args.vecfile, args.veclyr)
        i = 1
        for col in veccols:
            print("{0}: {1}".format(i, col))
            i = i + 1



"""
The zonal stats module provides functions to perform pixel-in-polygon or point-in-pixel analysis. 

For undertaking a pixel-in-polygon analysis you need to consider the size of the polygons with
respect to the size of the pixels being intersected. 

Where the pixels are small with respect to the polygons so there is at least one pixel within the polygon
then the best function to use is:

* rsgislib.zonalstats.calcZonalBandStats

If the pixels are large with respect to the polygons then use the following function which intersects
the polygon centroid.

* rsgislib.zonalstats.calcZonalPolyPtsBandStats

If the pixel size in between and/or polygons are varied in size such that it is not certain that all 
polygons will contain a pixel then the following function will first attempt to intersect the polygon 
with the pixels and if there is not a pixel within the polygon then the centriod is used.

* rsgislib.zonalstats.calcZonalBandStatsTestPolyPts

Alternatively, the other functions are slower to execute but have more options with respect to the
method of intersection. The options for intersection are:

* METHOD_POLYCONTAINSPIXEL = 0           # Polygon completely contains pixel
* METHOD_POLYCONTAINSPIXELCENTER = 1     # Pixel center is within the polygon
* METHOD_POLYOVERLAPSPIXEL = 2           # Polygon overlaps the pixel
* METHOD_POLYOVERLAPSORCONTAINSPIXEL = 3 # Polygon overlaps or contains the pixel
* METHOD_PIXELCONTAINSPOLY = 4           # Pixel contains the polygon
* METHOD_PIXELCONTAINSPOLYCENTER = 5     # Polygon center is within pixel
* METHOD_ADAPTIVE = 6                    # The method is chosen based on relative areas of pixel and polygon.
* METHOD_ENVELOPE = 7                    # All pixels in polygon envelope chosen
* METHOD_PIXELAREAINPOLY = 8             # Percent of pixel area that is within the polygon
* METHOD_POLYAREAINPIXEL = 9             # Percent of polygon area that is within pixel

"""
from __future__ import print_function

# import the C++ extension into this level
from ._zonalstats import *

import osgeo.gdal as gdal
import osgeo.ogr as ogr
import osgeo.osr as osr
import numpy
import math
import sys

METHOD_POLYCONTAINSPIXEL = 0           # Polygon completely contains pixel
METHOD_POLYCONTAINSPIXELCENTER = 1     # Pixel center is within the polygon
METHOD_POLYOVERLAPSPIXEL = 2           # Polygon overlaps the pixel
METHOD_POLYOVERLAPSORCONTAINSPIXEL = 3 # Polygon overlaps or contains the pixel
METHOD_PIXELCONTAINSPOLY = 4           # Pixel contains the polygon
METHOD_PIXELCONTAINSPOLYCENTER = 5     # Polygon center is within pixel
METHOD_ADAPTIVE = 6                    # The method is chosen based on relative areas of pixel and polygon.
METHOD_ENVELOPE = 7                    # All pixels in polygon envelope chosen
METHOD_PIXELAREAINPOLY = 8             # Percent of pixel area that is within the polygon
METHOD_POLYAREAINPIXEL = 9             # Percent of polygon area that is within pixel

class ZonalAttributes:
    """ Object, specifying which stats should be calculated and minimum / maximum thresholds. 
This is passed to the pixelStats2SHP and pixelStats2TXT functions. """
    def __init__(self, minThreshold=None, maxThreshold=None, calcCount=False, calcMin=False, calcMax=False, calcMean=False, calcStdDev=False, calcMode=False, calcSum=False):
        self.minThreshold = minThreshold
        self.maxThreshold = maxThreshold
        self.calcCount = calcCount
        self.calcMin = calcMin
        self.calcMax = calcMax
        self.calcMean = calcMean
        self.calcStdDev = calcStdDev 
        self.calcMode = calcMode
        self.calcSum = calcSum
        

class ZonalBandAttributes:
    """ Object, specifying which band, the band name and stats should be calculated and minimum / maximum thresholds. 
This is passed to the polyPixelStatsVecLyr function. """
    def __init__(self, band=0, basename="band", minThres=None, maxThres=None, calcCount=False, calcMin=False, calcMax=False, calcMean=False, calcStdDev=False, calcMode=False, calcMedian=False, calcSum=False):
        self.band = band
        self.basename = basename
        self.minThres = minThres
        self.maxThres = maxThres
        self.calcCount = calcCount
        self.calcMin = calcMin
        self.calcMax = calcMax
        self.calcMean = calcMean
        self.calcStdDev = calcStdDev 
        self.calcMode = calcMode
        self.calcMedian = calcMedian
        self.calcSum = calcSum

def calcZonalBandStatsFile(vecfile, veclyrname, valsimg, imgbandidx, minthres, maxthres, minfield=None, maxfield=None, meanfield=None, stddevfield=None, sumfield=None, countfield=None, modefield=None, medianfield=None):
    """
A function which calculates zonal statistics for a particular image band. 
If you know that the pixels in the values image are small with respect to 
the polygons then use this function.

:param vecfile: input vector file
:param veclyrname: input vector layer within the input file which specifies the features and where the output stats will be written.
:param valsimg: the values image
:param imgbandidx: the index (starting at 1) of the image band for which the stats will be calculated. If defined the no data value of the band will be ignored.
:param minthres: a lower threshold for values which will be included in the stats calculation.
:param maxthres: a upper threshold for values which will be included in the stats calculation.
:param minfield: the name of the field for the min value (None or not specified to be ignored).
:param maxfield: the name of the field for the max value (None or not specified to be ignored).
:param meanfield: the name of the field for the mean value (None or not specified to be ignored).
:param stddevfield: the name of the field for the standard deviation value (None or not specified to be ignored).
:param sumfield: the name of the field for the sum value (None or not specified to be ignored).
:param countfield: the name of the field for the count (of number of pixels) value (None or not specified to be ignored).
:param modefield: the name of the field for the mode value (None or not specified to be ignored).
:param medianfield: the name of the field for the median value (None or not specified to be ignored).

"""
    gdal.UseExceptions()
    
    try:    
        vecDS = gdal.OpenEx(vecfile, gdal.OF_VECTOR|gdal.OF_UPDATE )
        if vecDS is None:
            raise Exception("Could not open '{}'".format(vecfile)) 
    
        veclyr = vecDS.GetLayerByName(veclyrname)
        if veclyr is None:
            raise Exception("Could not open layer '{}'".format(veclyrname))
        veclyr_spatial_ref = veclyr.GetSpatialRef()
        
        calcZonalBandStats(veclyr, valsimg, imgbandidx, minthres, maxthres, minfield, maxfield, meanfield, stddevfield, sumfield, countfield, modefield, medianfield)
            
        vecDS = None
    except Exception as e:
        print("Error Vector File: {}".format(vecfile), file=sys.stderr)
        print("Error Vector Layer: {}".format(veclyrname), file=sys.stderr)
        print("Error Image File: {}".format(valsimg), file=sys.stderr)
        raise e


def calcZonalBandStats(veclyr, valsimg, imgbandidx, minthres, maxthres, minfield=None, maxfield=None, meanfield=None, stddevfield=None, sumfield=None, countfield=None, modefield=None, medianfield=None):
    """
A function which calculates zonal statistics for a particular image band. 
If you know that the pixels in the values image are small with respect to 
the polygons then use this function.

:param veclyr: OGR vector layer object containing the geometries being processed and to which the stats will be written.
:param valsimg: the values image
:param imgbandidx: the index (starting at 1) of the image band for which the stats will be calculated. If defined the no data value of the band will be ignored.
:param minthres: a lower threshold for values which will be included in the stats calculation.
:param maxthres: a upper threshold for values which will be included in the stats calculation.
:param minfield: the name of the field for the min value (None or not specified to be ignored).
:param maxfield: the name of the field for the max value (None or not specified to be ignored).
:param meanfield: the name of the field for the mean value (None or not specified to be ignored).
:param stddevfield: the name of the field for the standard deviation value (None or not specified to be ignored).
:param sumfield: the name of the field for the sum value (None or not specified to be ignored).
:param countfield: the name of the field for the count (of number of pixels) value (None or not specified to be ignored).
:param modefield: the name of the field for the mode value (None or not specified to be ignored).
:param medianfield: the name of the field for the median value (None or not specified to be ignored).

"""
    if modefield is not None:
        import scipy.stats.mstats
    gdal.UseExceptions()
    
    try:
        if veclyr is None:
            raise Exception("The inputted vector layer was None")
    
        if (minfield is None) and (maxfield is None) and (meanfield is None) and (stddevfield is None) and (sumfield is None) and (countfield is None) and (modefield is None) and (medianfield is None):
            raise Exception("At least one field needs to be specified for there is to an output.")
    
        imgDS = gdal.OpenEx(valsimg, gdal.GA_ReadOnly)
        if imgDS is None:
            raise Exception("Could not open '{}'".format(valsimg))
        imgband = imgDS.GetRasterBand(imgbandidx)
        if imgband is None:
            raise Exception("Could not find image band '{}'".format(imgbandidx))
        imgGeoTrans = imgDS.GetGeoTransform()
        img_wkt_str = imgDS.GetProjection()
        img_spatial_ref = osr.SpatialReference()
        img_spatial_ref.ImportFromWkt(img_wkt_str)
        
        pixel_width = imgGeoTrans[1]
        pixel_height = imgGeoTrans[5]
        
        imgSizeX = imgDS.RasterXSize
        imgSizeY = imgDS.RasterYSize
    
        imgNoDataVal = imgband.GetNoDataValue()

        veclyr_spatial_ref = veclyr.GetSpatialRef()
        
        if not img_spatial_ref.IsSame(veclyr_spatial_ref):
            imgDS = None
            vecDS = None
            raise Exception("Inputted raster and vector layers have different projections: ('{0}' '{1}') ".format(vecfile, valsimg))
        
        veclyrDefn = veclyr.GetLayerDefn()
        
        outFieldAtts = [minfield, maxfield, meanfield, stddevfield, sumfield, countfield, modefield, medianfield]
        for outattname in outFieldAtts:
            if outattname is not None:
                found = False
                for i in range(veclyrDefn.GetFieldCount()):
                    if veclyrDefn.GetFieldDefn(i).GetName().lower() == outattname.lower():
                        found = True
                        break
                if not found:
                    veclyr.CreateField(ogr.FieldDefn(outattname.lower(), ogr.OFTReal))
        
        fieldAttIdxs = dict()
        for outattname in outFieldAtts:
            if outattname is not None:
                fieldAttIdxs[outattname] = veclyr.FindFieldIndex(outattname.lower(), True)
        
        vec_mem_drv = ogr.GetDriverByName('Memory')
        img_mem_drv = gdal.GetDriverByName('MEM')
    
        # Iterate through features.
        openTransaction = False
        transactionStep = 20000
        nextTransaction = transactionStep
        nFeats = veclyr.GetFeatureCount(True)
        step = math.floor(nFeats/10)
        feedback = 10
        feedback_next = step
        counter = 0
        print("Started .0.", end='', flush=True)
        veclyr.ResetReading()
        feat = veclyr.GetNextFeature()
        while feat is not None:
            if (nFeats>10) and (counter == feedback_next):
                print(".{}.".format(feedback), end='', flush=True)
                feedback_next = feedback_next + step
                feedback = feedback + 10
            
            if not openTransaction:
                veclyr.StartTransaction()
                openTransaction = True
                
            if feat is not None:
                feat_geom = feat.geometry()
                if feat_geom is not None:
                    feat_bbox = feat_geom.GetEnvelope()
                    havepxls = True
                    
                    x1Sp = float(feat_bbox[0] - imgGeoTrans[0])
                    x2Sp = float(feat_bbox[1] - imgGeoTrans[0])
                    y1Sp = float(feat_bbox[3] - imgGeoTrans[3])
                    y2Sp = float(feat_bbox[2] - imgGeoTrans[3])
                                        
                    if x1Sp == 0.0:
                        x1 = 0
                    else:
                        x1 = int(x1Sp / pixel_width) - 1
                        
                    if x2Sp == 0.0:
                        x2 = 0
                    else:
                        x2 = int(x2Sp / pixel_width) + 1
                        
                    if y1Sp == 0.0:
                        y1 = 0
                    else:
                        y1 = int(y1Sp / pixel_height) - 1
                        
                    if y2Sp == 0.0:
                        y2 = 0
                    else:
                        y2 = int(y2Sp / pixel_height) + 1
                                        
                    if x1 < 0:
                        x1 = 0
                    elif x1 >= imgSizeX:
                        x1 = imgSizeX-1
                        
                    if x2 < 0:
                        x2 = 0
                    elif x2 >= imgSizeX:
                        x2 = imgSizeX-1
                        
                    if y1 < 0:
                        y1 = 0
                    elif y1 >= imgSizeY:
                        y1 = imgSizeY-1
                    
                    if y2 < 0:
                        y2 = 0
                    elif y2 >= imgSizeY:
                        y2 = imgSizeY-1
                                        
                    xsize = x2 - x1
                    ysize = y2 - y1
                    
                    if (xsize == 0) or (ysize == 0):
                        havepxls = False
                    
                    # Define the image ROI for the feature
                    src_offset = (x1, y1, xsize, ysize)

                    if havepxls:
                        # Read the band array.
                        src_array = imgband.ReadAsArray(*src_offset)
                    else:
                        src_array = None
                    
                    if (src_array is not None) and havepxls:
                
                        # calculate new geotransform of the feature subset
                        subGeoTrans = ((imgGeoTrans[0] + (src_offset[0] * imgGeoTrans[1])), imgGeoTrans[1], 0.0, (imgGeoTrans[3] + (src_offset[1] * imgGeoTrans[5])), 0.0, imgGeoTrans[5])
                
                        # Create a temporary vector layer in memory
                        vec_mem_ds = vec_mem_drv.CreateDataSource('out')
                        vec_mem_lyr = vec_mem_ds.CreateLayer('poly', veclyr_spatial_ref, ogr.wkbPolygon)
                        vec_mem_lyr.CreateFeature(feat.Clone())
                
                        # Rasterize the feature.
                        img_tmp_ds = img_mem_drv.Create('', src_offset[2], src_offset[3], 1, gdal.GDT_Byte)
                        img_tmp_ds.SetGeoTransform(subGeoTrans)
                        img_tmp_ds.SetProjection(img_wkt_str)
                        gdal.RasterizeLayer(img_tmp_ds, [1], vec_mem_lyr, burn_values=[1])
                        rv_array = img_tmp_ds.ReadAsArray()
                
                        # Mask the data vals array to feature (logical_not to flip 0<->1 to get the correct mask effect).
                        if imgNoDataVal is not None:
                            masked = numpy.ma.MaskedArray(src_array, mask=numpy.logical_or(src_array == imgNoDataVal, numpy.logical_not(rv_array), numpy.logical_and(src_array >= minthres, src_array < maxthres)))
                        else:
                            masked = numpy.ma.MaskedArray(src_array, mask=numpy.logical_or(numpy.logical_not(rv_array), numpy.logical_and(src_array >= minthres, src_array < maxthres)))
                        
                        if minfield is not None:
                            min_val = float(masked.min())
                            feat.SetField(fieldAttIdxs[minfield], min_val)
                        if maxfield is not None:
                            max_val = float(masked.max())
                            feat.SetField(fieldAttIdxs[maxfield], max_val)
                        if meanfield is not None:
                            mean_val = float(masked.mean())
                            feat.SetField(fieldAttIdxs[meanfield], mean_val)
                        if stddevfield is not None:
                            stddev_val = float(masked.std())
                            feat.SetField(fieldAttIdxs[stddevfield], stddev_val)
                        if sumfield is not None:
                            sum_val = float(masked.sum())
                            feat.SetField(fieldAttIdxs[sumfield], sum_val)
                        if countfield is not None:
                            count_val = float(masked.count())
                            feat.SetField(fieldAttIdxs[countfield], count_val)
                        if modefield is not None:
                            mode_val, mode_count = scipy.stats.mstats.mode(masked.flatten())
                            mode_val = float(mode_val)
                            feat.SetField(fieldAttIdxs[modefield], mode_val)
                        if medianfield is not None:
                            median_val = float(numpy.ma.median(masked))
                            feat.SetField(fieldAttIdxs[medianfield], median_val)
                        # Write the updated feature to the vector layer.
                        veclyr.SetFeature(feat)
            
                        vec_mem_ds = None
                        img_tmp_ds = None
            
            if (counter == nextTransaction) and openTransaction:
                veclyr.CommitTransaction()
                openTransaction = False
                nextTransaction = nextTransaction + transactionStep
            
            feat = veclyr.GetNextFeature()
            counter = counter + 1
        if openTransaction:
            veclyr.CommitTransaction()
            openTransaction = False
        print(" Completed")
    
        imgDS = None
    except Exception as e:
        print("Error Image File: {}".format(valsimg), file=sys.stderr)
        raise e


def calcZonalPolyPtsBandStatsFile(vecfile, veclyrname, valsimg, imgbandidx, outfield):
    """
A funtion which extracts zonal stats for a polygon using the polygon centroid.
This is useful when you are intesecting a low resolution image with respect to
the polygon resolution.

:param vecfile: input vector file
:param veclyrname: input vector layer within the input file which specifies the features and where the output stats will be written.
:param valsimg: the values image
:param imgbandidx: the index (starting at 1) of the image band for which the stats will be calculated. If defined the no data value of the band will be ignored.
:param outfield: output field name within the vector layer.

"""
    gdal.UseExceptions()
    try:    
        vecDS = gdal.OpenEx(vecfile, gdal.OF_VECTOR|gdal.OF_UPDATE )
        if vecDS is None:
            raise Exception("Could not open '{}'".format(vecfile)) 
    
        veclyr = vecDS.GetLayerByName(veclyrname)
        if veclyr is None:
            raise Exception("Could not open layer '{}'".format(veclyrname))
        veclyr_spatial_ref = veclyr.GetSpatialRef()
        
        calcZonalPolyPtsBandStats(veclyr, valsimg, imgbandidx, outfield)
            
        vecDS = None
    except Exception as e:
        print("Error Vector File: {}".format(vecfile), file=sys.stderr)
        print("Error Vector Layer: {}".format(veclyrname), file=sys.stderr)
        print("Error Image File: {}".format(valsimg), file=sys.stderr)
        raise e


def calcZonalPolyPtsBandStats(veclyr, valsimg, imgbandidx, outfield):
    """
A funtion which extracts zonal stats for a polygon using the polygon centroid.
This is useful when you are intesecting a low resolution image with respect to
the polygon resolution.

:param veclyr: OGR vector layer object containing the geometries being processed and to which the stats will be written.
:param valsimg: the values image
:param imgbandidx: the index (starting at 1) of the image band for which the stats will be calculated. If defined the no data value of the band will be ignored.
:param outfield: output field name within the vector layer.

"""
    gdal.UseExceptions()
    try:
        if veclyr is None:
            raise Exception("The inputted vector layer was None")
    
        imgDS = gdal.OpenEx(valsimg, gdal.GA_ReadOnly)
        if imgDS is None:
            raise Exception("Could not open '{}'".format(valsimg))
        imgband = imgDS.GetRasterBand(imgbandidx)
        if imgband is None:
            raise Exception("Could not find image band '{}'".format(imgbandidx))
        imgGeoTrans = imgDS.GetGeoTransform()
        img_wkt_str = imgDS.GetProjection()
        img_spatial_ref = osr.SpatialReference()
        img_spatial_ref.ImportFromWkt(img_wkt_str)
        
        pixel_width = imgGeoTrans[1]
        pixel_height = imgGeoTrans[5]
        
        imgSizeX = imgDS.RasterXSize
        imgSizeY = imgDS.RasterYSize
    
        veclyr_spatial_ref = veclyr.GetSpatialRef()
        
        if not img_spatial_ref.IsSame(veclyr_spatial_ref):
            imgDS = None
            vecDS = None
            raise Exception("Inputted raster and vector layers have different projections: ('{0}' '{1}') ".format(vecfile, valsimg))
        
        veclyrDefn = veclyr.GetLayerDefn()
        
        found = False
        for i in range(veclyrDefn.GetFieldCount()):
            if veclyrDefn.GetFieldDefn(i).GetName().lower() == outfield.lower():
                found = True
                break
        if not found:
            veclyr.CreateField(ogr.FieldDefn(outfield.lower(), ogr.OFTReal))
        
        outfieldidx = veclyr.FindFieldIndex(outfield.lower(), True)
    
        vec_mem_drv = ogr.GetDriverByName('Memory')
        img_mem_drv = gdal.GetDriverByName('MEM')
        
        # Iterate through features.
        openTransaction = False
        transactionStep = 20000
        nextTransaction = transactionStep
        nFeats = veclyr.GetFeatureCount(True)
        step = math.floor(nFeats/10)
        feedback = 10
        feedback_next = step
        counter = 0
        print("Started .0.", end='', flush=True)
        veclyr.ResetReading()
        feat = veclyr.GetNextFeature()
        while feat is not None:
            if (nFeats>10) and (counter == feedback_next):
                print(".{}.".format(feedback), end='', flush=True)
                feedback_next = feedback_next + step
                feedback = feedback + 10
            
            if not openTransaction:
                veclyr.StartTransaction()
                openTransaction = True
            
            if feat is not None:
                feat_geom = feat.geometry()
                if feat_geom is not None:
                    feat_bbox = feat_geom.GetEnvelope()
                    havepxls = True
                    
                    x1Sp = float(feat_bbox[0] - imgGeoTrans[0])
                    x2Sp = float(feat_bbox[1] - imgGeoTrans[0])
                    y1Sp = float(feat_bbox[3] - imgGeoTrans[3])
                    y2Sp = float(feat_bbox[2] - imgGeoTrans[3])
                                        
                    if x1Sp == 0.0:
                        x1 = 0
                    else:
                        x1 = int(x1Sp / pixel_width) - 1
                        
                    if x2Sp == 0.0:
                        x2 = 0
                    else:
                        x2 = int(x2Sp / pixel_width) + 1
                        
                    if y1Sp == 0.0:
                        y1 = 0
                    else:
                        y1 = int(y1Sp / pixel_height) - 1
                        
                    if y2Sp == 0.0:
                        y2 = 0
                    else:
                        y2 = int(y2Sp / pixel_height) + 1
                                        
                    if x1 < 0:
                        x1 = 0
                    elif x1 >= imgSizeX:
                        x1 = imgSizeX-1
                        
                    if x2 < 0:
                        x2 = 0
                    elif x2 >= imgSizeX:
                        x2 = imgSizeX-1
                        
                    if y1 < 0:
                        y1 = 0
                    elif y1 >= imgSizeY:
                        y1 = imgSizeY-1
                    
                    if y2 < 0:
                        y2 = 0
                    elif y2 >= imgSizeY:
                        y2 = imgSizeY-1
                                        
                    xsize = x2 - x1
                    ysize = y2 - y1
                    
                    if (xsize == 0) or (ysize == 0):
                        havepxls = False
                    
                    # Define the image ROI for the feature
                    src_offset = (x1, y1, xsize, ysize)

                    if havepxls:
                        # Read the band array.
                        src_array = imgband.ReadAsArray(*src_offset)
                    else:
                        src_array = None
                    
                    if (src_array is not None) and havepxls:
                        subTLX = (imgGeoTrans[0] + (src_offset[0] * imgGeoTrans[1]))
                        subTLY = (imgGeoTrans[3] + (src_offset[1] * imgGeoTrans[5]))
                        resX = imgGeoTrans[1]
                        resY = imgGeoTrans[5]
                                    
                        ptx,pty,ptz = feat.GetGeometryRef().Centroid().GetPoint()
                        
                        xOff = math.floor((ptx - subTLX) / resX)
                        yOff = math.floor((pty - subTLY) / resY)
                        
                        if xOff < 0:
                            xOff = 0
                        if xOff >= xsize:
                            xOff = xsize - 1
                            
                        if yOff < 0:
                            yOff = 0
                        if yOff >= ysize:
                            yOff = ysize - 1
                
                        out_val = float(src_array[yOff, xOff])
                        feat.SetField(outfieldidx, out_val)
            
                        veclyr.SetFeature(feat)
            
                        vec_mem_ds = None
                        img_tmp_ds = None
            
            if (counter == nextTransaction) and openTransaction:
                veclyr.CommitTransaction()
                openTransaction = False
                nextTransaction = nextTransaction + transactionStep
            
            feat = veclyr.GetNextFeature()
            counter = counter + 1
    
        if openTransaction:
            veclyr.CommitTransaction()
            openTransaction = False
        print(" Completed")
    
        imgDS = None
    except Exception as e:
        print("Error Image File: {}".format(valsimg), file=sys.stderr)
        raise e


def calcZonalBandStatsTestPolyPtsFile(vecfile, veclyrname, valsimg, imgbandidx, minthres, maxthres, minfield=None, maxfield=None, meanfield=None, stddevfield=None, sumfield=None, countfield=None, modefield=None, medianfield=None):
    """
A function which calculates zonal statistics for a particular image band. If unsure then use this function. 
This function tests whether 1 or more pixels has been found within the polygon and if not then the centroid 
use used to find a value for the polygon.

If you are unsure as to whether the pixels are small enough to be contained within all the polygons then 
use this function.

:param vecfile: input vector file
:param veclyrname: input vector layer within the input file which specifies the features and where the output stats will be written.
:param valsimg: the values image
:param imgbandidx: the index (starting at 1) of the image band for which the stats will be calculated. If defined the no data value of the band will be ignored.
:param minthres: a lower threshold for values which will be included in the stats calculation.
:param maxthres: a upper threshold for values which will be included in the stats calculation.
:param minfield: the name of the field for the min value (None or not specified to be ignored).
:param maxfield: the name of the field for the max value (None or not specified to be ignored).
:param meanfield: the name of the field for the mean value (None or not specified to be ignored).
:param stddevfield: the name of the field for the standard deviation value (None or not specified to be ignored).
:param sumfield: the name of the field for the sum value (None or not specified to be ignored).
:param countfield: the name of the field for the count (of number of pixels) value (None or not specified to be ignored).
:param modefield: the name of the field for the mode value (None or not specified to be ignored).
:param medianfield: the name of the field for the median value (None or not specified to be ignored).

"""
    gdal.UseExceptions()
    
    try:    
        vecDS = gdal.OpenEx(vecfile, gdal.OF_VECTOR|gdal.OF_UPDATE )
        if vecDS is None:
            raise Exception("Could not open '{}'".format(vecfile)) 
    
        veclyr = vecDS.GetLayerByName(veclyrname)
        if veclyr is None:
            raise Exception("Could not open layer '{}'".format(veclyrname))
            
        calcZonalBandStatsTestPolyPts(veclyr, valsimg, imgbandidx, minthres, maxthres, minfield, maxfield, meanfield, stddevfield, sumfield, countfield, modefield, medianfield)
    
        vecDS = None
    except Exception as e:
        print("Error Vector File: {}".format(vecfile), file=sys.stderr)
        print("Error Vector Layer: {}".format(veclyrname), file=sys.stderr)
        print("Error Image File: {}".format(valsimg), file=sys.stderr)
        raise e


def calcZonalBandStatsTestPolyPts(veclyr, valsimg, imgbandidx, minthres, maxthres, minfield=None, maxfield=None, meanfield=None, stddevfield=None, sumfield=None, countfield=None, modefield=None, medianfield=None):
    """
A function which calculates zonal statistics for a particular image band. If unsure then use this function. 
This function tests whether 1 or more pixels has been found within the polygon and if not then the centroid 
use used to find a value for the polygon.

If you are unsure as to whether the pixels are small enough to be contained within all the polygons then 
use this function.

:param veclyr: OGR vector layer object containing the geometries being processed and to which the stats will be written.
:param valsimg: the values image
:param imgbandidx: the index (starting at 1) of the image band for which the stats will be calculated. If defined the no data value of the band will be ignored.
:param minthres: a lower threshold for values which will be included in the stats calculation.
:param maxthres: a upper threshold for values which will be included in the stats calculation.
:param minfield: the name of the field for the min value (None or not specified to be ignored).
:param maxfield: the name of the field for the max value (None or not specified to be ignored).
:param meanfield: the name of the field for the mean value (None or not specified to be ignored).
:param stddevfield: the name of the field for the standard deviation value (None or not specified to be ignored).
:param sumfield: the name of the field for the sum value (None or not specified to be ignored).
:param countfield: the name of the field for the count (of number of pixels) value (None or not specified to be ignored).
:param modefield: the name of the field for the mode value (None or not specified to be ignored).
:param medianfield: the name of the field for the median value (None or not specified to be ignored).

"""
    if modefield is not None:
        import scipy.stats.mstats
    gdal.UseExceptions()
    
    try:
        if veclyr is None:
            raise Exception("The inputted vector layer was None")
    
        if (minfield is None) and (maxfield is None) and (meanfield is None) and (stddevfield is None) and (sumfield is None) and (countfield is None) and (modefield is None) and (medianfield is None):
            raise Exception("At least one field needs to be specified for there is to an output.")
    
        imgDS = gdal.OpenEx(valsimg, gdal.GA_ReadOnly)
        if imgDS is None:
            raise Exception("Could not open '{}'".format(valsimg))
        imgband = imgDS.GetRasterBand(imgbandidx)
        if imgband is None:
            raise Exception("Could not find image band '{}'".format(imgbandidx))
        imgGeoTrans = imgDS.GetGeoTransform()
        img_wkt_str = imgDS.GetProjection()
        img_spatial_ref = osr.SpatialReference()
        img_spatial_ref.ImportFromWkt(img_wkt_str)
        
        pixel_width = imgGeoTrans[1]
        pixel_height = imgGeoTrans[5]
        
        imgSizeX = imgDS.RasterXSize
        imgSizeY = imgDS.RasterYSize
            
        imgNoDataVal = imgband.GetNoDataValue()
    
        veclyr_spatial_ref = veclyr.GetSpatialRef()
        
        if not img_spatial_ref.IsSame(veclyr_spatial_ref):
            imgDS = None
            vecDS = None
            raise Exception("Inputted raster and vector layers have different projections: ('{0}' '{1}') ".format(vecfile, valsimg))
        
        veclyrDefn = veclyr.GetLayerDefn()
        
        outFieldAtts = [minfield, maxfield, meanfield, stddevfield, sumfield, countfield, modefield, medianfield]
        for outattname in outFieldAtts:
            if outattname is not None:
                found = False
                for i in range(veclyrDefn.GetFieldCount()):
                    if veclyrDefn.GetFieldDefn(i).GetName().lower() == outattname.lower():
                        found = True
                        break
                if not found:
                    veclyr.CreateField(ogr.FieldDefn(outattname.lower(), ogr.OFTReal))
        
        fieldAttIdxs = dict()
        for outattname in outFieldAtts:
            if outattname is not None:
                fieldAttIdxs[outattname] = veclyr.FindFieldIndex(outattname.lower(), True)
        
        vec_mem_drv = ogr.GetDriverByName('Memory')
        img_mem_drv = gdal.GetDriverByName('MEM')
    
        # Iterate through features.
        openTransaction = False
        transactionStep = 20000
        nextTransaction = transactionStep
        nFeats = veclyr.GetFeatureCount(True)
        step = math.floor(nFeats/10)
        feedback = 10
        feedback_next = step
        counter = 0
        print("Started .0.", end='', flush=True)
        veclyr.ResetReading()
        feat = veclyr.GetNextFeature()
        while feat is not None:
            if (nFeats>10) and (counter == feedback_next):
                print(".{}.".format(feedback), end='', flush=True)
                feedback_next = feedback_next + step
                feedback = feedback + 10
            
            if not openTransaction:
                veclyr.StartTransaction()
                openTransaction = True
            
            if feat is not None:
                feat_geom = feat.geometry()
                if feat_geom is not None:
                    feat_bbox = feat_geom.GetEnvelope()
                    
                    havepxls = True
                    
                    x1Sp = float(feat_bbox[0] - imgGeoTrans[0])
                    x2Sp = float(feat_bbox[1] - imgGeoTrans[0])
                    y1Sp = float(feat_bbox[3] - imgGeoTrans[3])
                    y2Sp = float(feat_bbox[2] - imgGeoTrans[3])
                                        
                    if x1Sp == 0.0:
                        x1 = 0
                    else:
                        x1 = int(x1Sp / pixel_width) - 1
                        
                    if x2Sp == 0.0:
                        x2 = 0
                    else:
                        x2 = int(x2Sp / pixel_width) + 1
                        
                    if y1Sp == 0.0:
                        y1 = 0
                    else:
                        y1 = int(y1Sp / pixel_height) - 1
                        
                    if y2Sp == 0.0:
                        y2 = 0
                    else:
                        y2 = int(y2Sp / pixel_height) + 1
                                        
                    if x1 < 0:
                        x1 = 0
                    elif x1 >= imgSizeX:
                        x1 = imgSizeX-1
                        
                    if x2 < 0:
                        x2 = 0
                    elif x2 >= imgSizeX:
                        x2 = imgSizeX-1
                        
                    if y1 < 0:
                        y1 = 0
                    elif y1 >= imgSizeY:
                        y1 = imgSizeY-1
                    
                    if y2 < 0:
                        y2 = 0
                    elif y2 >= imgSizeY:
                        y2 = imgSizeY-1
                                        
                    xsize = x2 - x1
                    ysize = y2 - y1
                    
                    if (xsize == 0) or (ysize == 0):
                        havepxls = False
                    
                    # Define the image ROI for the feature
                    src_offset = (x1, y1, xsize, ysize)

                    if havepxls:
                        # Read the band array.
                        src_array = imgband.ReadAsArray(*src_offset)
                    else:
                        src_array = None
                    
                    if (src_array is not None) and havepxls:
                
                        # calculate new geotransform of the feature subset
                        subGeoTrans = ((imgGeoTrans[0] + (src_offset[0] * imgGeoTrans[1])), imgGeoTrans[1], 0.0, (imgGeoTrans[3] + (src_offset[1] * imgGeoTrans[5])), 0.0, imgGeoTrans[5])
                
                        # Create a temporary vector layer in memory
                        vec_mem_ds = vec_mem_drv.CreateDataSource('out')
                        vec_mem_lyr = vec_mem_ds.CreateLayer('poly', veclyr_spatial_ref, ogr.wkbPolygon)
                        vec_mem_lyr.CreateFeature(feat.Clone())
                
                        # Rasterize the feature.
                        img_tmp_ds = img_mem_drv.Create('', src_offset[2], src_offset[3], 1, gdal.GDT_Byte)
                        img_tmp_ds.SetGeoTransform(subGeoTrans)
                        img_tmp_ds.SetProjection(img_wkt_str)
                        gdal.RasterizeLayer(img_tmp_ds, [1], vec_mem_lyr, burn_values=[1])
                        rv_array = img_tmp_ds.ReadAsArray()
                
                        # Mask the data vals array to feature (logical_not to flip 0<->1 to get the correct mask effect).
                        if imgNoDataVal is not None:
                            masked = numpy.ma.MaskedArray(src_array, mask=numpy.logical_or(src_array == imgNoDataVal, numpy.logical_not(rv_array), numpy.logical_and(src_array >= minthres, src_array < maxthres)))
                        else:
                            masked = numpy.ma.MaskedArray(src_array, mask=numpy.logical_or(numpy.logical_not(rv_array), numpy.logical_and(src_array >= minthres, src_array < maxthres)))
                        
                        if float(masked.count()) > 0:
                            if minfield is not None:
                                min_val = float(masked.min())
                                feat.SetField(fieldAttIdxs[minfield], min_val)
                            if maxfield is not None:
                                max_val = float(masked.max())
                                feat.SetField(fieldAttIdxs[maxfield], max_val)
                            if meanfield is not None:
                                mean_val = float(masked.mean())
                                feat.SetField(fieldAttIdxs[meanfield], mean_val)
                            if stddevfield is not None:
                                stddev_val = float(masked.std())
                                feat.SetField(fieldAttIdxs[stddevfield], stddev_val)
                            if sumfield is not None:
                                sum_val = float(masked.sum())
                                feat.SetField(fieldAttIdxs[sumfield], sum_val)
                            if countfield is not None:
                                count_val = float(masked.count())
                                feat.SetField(fieldAttIdxs[countfield], count_val)
                            if modefield is not None:
                                mode_val, mode_count = scipy.stats.mstats.mode(masked.flatten())
                                mode_val = float(mode_val)
                                feat.SetField(fieldAttIdxs[modefield], mode_val)
                            if medianfield is not None:
                                median_val = float(numpy.ma.median(masked))
                                feat.SetField(fieldAttIdxs[medianfield], median_val)
                        else:
                            subTLX = (imgGeoTrans[0] + (src_offset[0] * imgGeoTrans[1]))
                            subTLY = (imgGeoTrans[3] + (src_offset[1] * imgGeoTrans[5]))
                            resX = imgGeoTrans[1]
                            resY = imgGeoTrans[5]
                                        
                            ptx,pty,ptz = feat.GetGeometryRef().Centroid().GetPoint()
                            
                            xOff = math.floor((ptx - subTLX) / resX)
                            yOff = math.floor((pty - subTLY) / resY)
                            
                            if xOff < 0:
                                xOff = 0
                            if xOff >= xsize:
                                xOff = xsize - 1
                                
                            if yOff < 0:
                                yOff = 0
                            if yOff >= ysize:
                                yOff = ysize - 1
                            
                            out_val = float(src_array[yOff, xOff])
                            if minfield is not None:
                                feat.SetField(fieldAttIdxs[minfield], out_val)
                            if maxfield is not None:
                                feat.SetField(fieldAttIdxs[maxfield], out_val)
                            if meanfield is not None:
                                feat.SetField(fieldAttIdxs[meanfield], out_val)
                            if stddevfield is not None:
                                feat.SetField(fieldAttIdxs[stddevfield], 0.0)
                            if sumfield is not None:
                                feat.SetField(fieldAttIdxs[sumfield], out_val)
                            if countfield is not None:
                                feat.SetField(fieldAttIdxs[countfield], 1.0)
                            if modefield is not None:
                                feat.SetField(fieldAttIdxs[modefield], out_val)
                            if medianfield is not None:
                                feat.SetField(fieldAttIdxs[medianfield], out_val)
                            
                        # Write the updated feature to the vector layer.
                        veclyr.SetFeature(feat)
            
                        vec_mem_ds = None
                        img_tmp_ds = None            
            
            if (counter == nextTransaction) and openTransaction:
                veclyr.CommitTransaction()
                openTransaction = False
                nextTransaction = nextTransaction + transactionStep
            
            feat = veclyr.GetNextFeature()
            counter = counter + 1
        if openTransaction:
            veclyr.CommitTransaction()
            openTransaction = False
        print(" Completed")
    
        imgDS = None
    except Exception as e:
        print("Error Image File: {}".format(valsimg), file=sys.stderr)
        raise e







import rsgislib
import rsgislib.imageutils
import rsgislib.imagecalc
import rsgislib.rastergis

from osgeo import gdal

import numpy

import skimage.segmentation

import os.path


def performFelsenszwalbSegmentation(inputImg, outputImg, gdalformat='KEA', noDataVal=0, tmpDIR='./tmp', calcStats=True, usePCA=False, nPCABands=3, pcaPxlSample=100, scale=1, sigma=0.8, min_size=20):
    """
A function to perform the Felsenszwalb segmentation algorithm from the
scikit-image library (http://scikit-image.org/docs/stable/api/skimage.segmentation.html).

:param inputImg: input image file.
:param outputImg: output image file.
:param gdalformat: output image file format.
:param tmpDIR: temp DIR used to output PCA files
:param calcStats: calculate image pixel statistics, histogram and image pyramids - note if you are not using a KEA file then the format needs to support RATs for this option as histogram and colour table are written to RAT.
:param usePCA: if there are not 1 or 3 image bands in the input file then you can use PCA to reduce the number of image bands.
:param nPCABands: the number of principle components outputs from the PCA - needs to be either 1 or 3.
:param scale: scikit-image Felsenszwalb parameter: 'Free parameter. Higher means larger clusters.'
:param sigma: scikit-image Felsenszwalb parameter: 'Width of Gaussian kernel used in preprocessing.'
:param min_size: scikit-image Felsenszwalb parameter: 'Minimum component size. Enforced using postprocessing.'

"""
    # Create output image
    rsgislib.imageutils.createCopyImage(inputImg, outputImg, 1, 0, gdalformat, rsgislib.TYPE_32UINT)
            
    gdalDS = gdal.Open(inputImg, gdal.GA_ReadOnly)
    pcaImg = ''
    if usePCA and (nPCABands == 1) and (gdalDS.RasterCount > 1):
        # Perform PCA
        gdalDS = None
        rsgisUtils = rsgislib.RSGISPyUtils()
        inImgBaseName = os.path.splitext(os.path.basename(inputImg))[0]
        pcaImg = os.path.join(tmpDIR, inImgBaseName+'_pca.kea')
        pcaEigenValsFile = os.path.join(tmpDIR, inImgBaseName+'_eigen.mtxt')
        inDType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImg)
        rsgislib.imagecalc.performImagePCA(inputImg, pcaImg, pcaEigenValsFile, 1, pcaPxlSample, 'KEA', inDType, noDataVal, False)
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif usePCA and (nPCABands == 3) and (gdalDS.RasterCount > 3):
        # Perform PCA
        gdalDS = None
        rsgisUtils = rsgislib.RSGISPyUtils()
        inImgBaseName = os.path.splitext(os.path.basename(inputImg))[0]
        pcaImg = os.path.join(tmpDIR, inImgBaseName+'_pca.kea')
        pcaEigenValsFile = os.path.join(tmpDIR, inImgBaseName+'_eigen.mtxt')
        inDType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImg)
        rsgislib.imagecalc.performImagePCA(inputImg, pcaImg, pcaEigenValsFile, 3, pcaPxlSample, 'KEA', inDType, noDataVal, False)
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif usePCA:
        gdalDS = None
        raise Exception('In the number of bands given the number of principle components.')    
    
    if not ((gdalDS.RasterCount == 1) or (gdalDS.RasterCount == 3)):
        gdalDS = None
        raise Exception('Input image should have either a single band or three (RGB). Consider performing PCA to reduce.')
    
    pxlValsArr = []
    multichannel=True
    if gdalDS.RasterCount > 1:
        for nBand in numpy.arange(gdalDS.RasterCount):
            gdalBand = gdalDS.GetRasterBand(int(nBand+1))
            pxlValsArr.append(gdalBand.ReadAsArray())
        pxlVals = numpy.stack(pxlValsArr, axis=-1)
        multichannel=True
    else:
        gdalBand = gdalDS.GetRasterBand(1)
        pxlVals = gdalBand.ReadAsArray()
        multichannel=False
    
    segResult = skimage.segmentation.felzenszwalb(pxlVals, scale=scale, sigma=sigma, min_size=min_size, multichannel=multichannel)
    segResult = segResult + 1
    
    gdalOutDS = gdal.Open(outputImg, gdal.GA_Update)
    gdalOutBand = gdalOutDS.GetRasterBand(1)
    gdalOutBand.WriteArray(segResult)
    gdalOutDS = None
    gdalDS = None
    
    if calcStats:
        rsgislib.rastergis.populateStats(clumps=outputImg, addclrtab=True, calcpyramids=True, ignorezero=True)
        
        
def performQuickshiftSegmentation(inputImg, outputImg, gdalformat='KEA', noDataVal=0, tmpDIR='./tmp', calcStats=True, usePCA=False, pcaPxlSample=100, ratio=1.0, kernel_size=5, max_dist=10, sigma=0, convert2lab=True, random_seed=42):
    """
A function to perform the quickshift segmentation algorithm from the
scikit-image library (http://scikit-image.org/docs/stable/api/skimage.segmentation.html).

:param inputImg: input image file.
:param outputImg: output image file.
:param gdalformat: output image file format.
:param tmpDIR: temp DIR used to output PCA files
:param calcStats: calculate image pixel statistics, histogram and image pyramids - note if you are not using a KEA file then the format needs to support RATs for this option as histogram and colour table are written to RAT.
:param usePCA: if there are not 3 image bands in the input file then you can use PCA to reduce the number of image bands.
:param ratio: scikit-image Quickshift parameter: 'Balances color-space proximity and image-space proximity. Higher values give more weight to color-space. (between 0 and 1)'
:param kernel_size: scikit-image Quickshift parameter: 'Width of Gaussian kernel used in smoothing the sample density. Higher means fewer clusters.'
:param max_dist: scikit-image Quickshift parameter: 'Cut-off point for data distances. Higher means fewer clusters.'
:param sigma: scikit-image Quickshift parameter: 'Width for Gaussian smoothing as preprocessing. Zero means no smoothing.'
:param convert2lab: scikit-image Quickshift parameter: 'Whether the input should be converted to Lab colorspace prior to segmentation. For this purpose, the input is assumed to be RGB.'
:param random_seed: scikit-image Quickshift parameter: 'Random seed used for breaking ties.'

"""
    # Create output image
    rsgislib.imageutils.createCopyImage(inputImg, outputImg, 1, 0, gdalformat, rsgislib.TYPE_32UINT)
            
    gdalDS = gdal.Open(inputImg, gdal.GA_ReadOnly)
    pcaImg = ''
    if usePCA and (gdalDS.RasterCount > 3):
        # Perform PCA
        gdalDS = None
        rsgisUtils = rsgislib.RSGISPyUtils()
        inImgBaseName = os.path.splitext(os.path.basename(inputImg))[0]
        pcaImg = os.path.join(tmpDIR, inImgBaseName+'_pca.kea')
        pcaEigenValsFile = os.path.join(tmpDIR, inImgBaseName+'_eigen.mtxt')
        inDType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImg)
        rsgislib.imagecalc.performImagePCA(inputImg, pcaImg, pcaEigenValsFile, 3, pcaPxlSample, 'KEA', inDType, noDataVal, False)
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif usePCA:
        gdalDS = None
        raise Exception('In the number of bands given the number of principle components.')
    
    if not (gdalDS.RasterCount == 3):
        gdalDS = None
        raise Exception('Input image should have three bands (RGB). Consider performing PCA to reduce.')
    
    pxlValsArr = []
    if gdalDS.RasterCount > 1:
        for nBand in numpy.arange(gdalDS.RasterCount):
            gdalBand = gdalDS.GetRasterBand(int(nBand+1))
            pxlValsArr.append(gdalBand.ReadAsArray())
        pxlVals = numpy.stack(pxlValsArr, axis=-1)
    else:
        gdalBand = gdalDS.GetRasterBand(1)
        pxlVals = gdalBand.ReadAsArray()
    
    segResult = skimage.segmentation.quickshift(pxlVals, ratio=ratio, kernel_size=kernel_size, max_dist=max_dist, return_tree=False, sigma=sigma, convert2lab=convert2lab, random_seed=random_seed)
    segResult = segResult + 1
    
    gdalOutDS = gdal.Open(outputImg, gdal.GA_Update)
    gdalOutBand = gdalOutDS.GetRasterBand(1)
    gdalOutBand.WriteArray(segResult)
    gdalOutDS = None
    gdalDS = None
    
    if calcStats:
        rsgislib.rastergis.populateStats(clumps=outputImg, addclrtab=True, calcpyramids=True, ignorezero=True)

def performSlicSegmentation(inputImg, outputImg, gdalformat='KEA', noDataVal=0, tmpDIR='./tmp', calcStats=True, usePCA=False, nPCABands=3, pcaPxlSample=100, n_segments=100, compactness=10.0, max_iter=10, sigma=0, spacing=None, convert2lab=None, enforce_connectivity=True, min_size_factor=0.5, max_size_factor=3, slic_zero=False):
    """
A function to perform the slic segmentation algorithm from the
scikit-image library (http://scikit-image.org/docs/stable/api/skimage.segmentation.html).

:param inputImg: input image file.
:param outputImg: output image file.
:param gdalformat: output image file format.
:param tmpDIR: temp DIR used to output PCA files
:param calcStats: calculate image pixel statistics, histogram and image pyramids - note if you are not using a KEA file then the format needs to support RATs for this option as histogram and colour table are written to RAT.
:param usePCA: if there are not 1 or 3 image bands in the input file then you can use PCA to reduce the number of image bands.
:param nPCABands: the number of principle components outputs from the PCA - needs to be either 1 or 3.
:param n_segments: scikit-image Slic parameter: 'The (approximate) number of labels in the segmented output image.'
:param compactness: scikit-image Slic parameter: 'Balances color proximity and space proximity. Higher values give more weight to space proximity, making superpixel shapes more square/cubic. In SLICO mode, this is the initial compactness. This parameter depends strongly on image contrast and on the shapes of objects in the image. We recommend exploring possible values on a log scale, e.g., 0.01, 0.1, 1, 10, 100, before refining around a chosen value.'
:param max_iter: scikit-image Slic parameter: 'Maximum number of iterations of k-means.'
:param sigma: scikit-image Slic parameter: 'Width of Gaussian smoothing kernel for pre-processing for each dimension of the image. The same sigma is applied to each dimension in case of a scalar value. Zero means no smoothing. Note, that sigma is automatically scaled if it is scalar and a manual voxel spacing is provided (see Notes section).'
:param spacing: scikit-image Slic parameter: 'The voxel spacing along each image dimension. By default, slic assumes uniform spacing (same voxel resolution along z, y and x). This parameter controls the weights of the distances along z, y, and x during k-means clustering.'
:param convert2lab: scikit-image Slic parameter: 'Whether the input should be converted to Lab colorspace prior to segmentation. The input image must be RGB. Highly recommended.'
:param enforce_connectivity: scikit-image Slic parameter: 'Whether the generated segments are connected or not'
:param min_size_factor: scikit-image Slic parameter: 'Proportion of the minimum segment size to be removed with respect to the supposed segment size "depth:paramwidth*height/n_segments"'
:param max_size_factor: scikit-image Slic parameter: 'Proportion of the maximum connected segment size. A value of 3 works in most of the cases.'
:param slic_zero: scikit-image Slic parameter: 'Run SLIC-zero, the zero-parameter mode of SLIC.'

"""
    # Create output image
    rsgislib.imageutils.createCopyImage(inputImg, outputImg, 1, 0, gdalformat, rsgislib.TYPE_32UINT)
            
    gdalDS = gdal.Open(inputImg, gdal.GA_ReadOnly)
    pcaImg = ''
    if usePCA and (nPCABands == 1) and (gdalDS.RasterCount > 1):
        # Perform PCA
        gdalDS = None
        rsgisUtils = rsgislib.RSGISPyUtils()
        inImgBaseName = os.path.splitext(os.path.basename(inputImg))[0]
        pcaImg = os.path.join(tmpDIR, inImgBaseName+'_pca.kea')
        pcaEigenValsFile = os.path.join(tmpDIR, inImgBaseName+'_eigen.mtxt')
        inDType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImg)
        rsgislib.imagecalc.performImagePCA(inputImg, pcaImg, pcaEigenValsFile, 1, pcaPxlSample, 'KEA', inDType, noDataVal, False)
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif usePCA and (nPCABands == 3) and (gdalDS.RasterCount > 3):
        # Perform PCA
        gdalDS = None
        rsgisUtils = rsgislib.RSGISPyUtils()
        inImgBaseName = os.path.splitext(os.path.basename(inputImg))[0]
        pcaImg = os.path.join(tmpDIR, inImgBaseName+'_pca.kea')
        pcaEigenValsFile = os.path.join(tmpDIR, inImgBaseName+'_eigen.mtxt')
        inDType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImg)
        rsgislib.imagecalc.performImagePCA(inputImg, pcaImg, pcaEigenValsFile, 3, pcaPxlSample, 'KEA', inDType, noDataVal, False)
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif usePCA:
        gdalDS = None
        raise Exception('In the number of bands given the number of principle components.')   
    
    if not ((gdalDS.RasterCount == 1) or (gdalDS.RasterCount == 3)):
        gdalDS = None
        raise Exception('Input image should have either a single band or three (RGB). Consider performing PCA to reduce.')
    
    pxlValsArr = []
    multichannel = False
    if gdalDS.RasterCount > 1:
        for nBand in numpy.arange(gdalDS.RasterCount):
            gdalBand = gdalDS.GetRasterBand(int(nBand+1))
            pxlValsArr.append(gdalBand.ReadAsArray())
        pxlVals = numpy.stack(pxlValsArr, axis=-1)
        multichannel = True
    else:
        gdalBand = gdalDS.GetRasterBand(1)
        pxlVals = gdalBand.ReadAsArray()
        multichannel = False
    
    segResult = skimage.segmentation.slic(pxlVals, n_segments=n_segments, compactness=compactness, max_iter=max_iter, sigma=sigma, spacing=spacing, multichannel=multichannel, convert2lab=convert2lab, enforce_connectivity=enforce_connectivity, min_size_factor=min_size_factor, max_size_factor=max_size_factor, slic_zero=slic_zero)
    segResult = segResult + 1
    
    gdalOutDS = gdal.Open(outputImg, gdal.GA_Update)
    gdalOutBand = gdalOutDS.GetRasterBand(1)
    gdalOutBand.WriteArray(segResult)
    gdalOutDS = None
    gdalDS = None
    
    if calcStats:
        rsgislib.rastergis.populateStats(clumps=outputImg, addclrtab=True, calcpyramids=True, ignorezero=True)


def performWatershedSegmentation(inputImg, markersImg, outputImg, gdalformat='KEA', noDataVal=0, tmpDIR='./tmp', calcStats=True, usePCA=False, nPCABands=3, pcaPxlSample=100, compactness=0, watershed_line=False):
    """
A function to perform the watershed segmentation algorithm from the
scikit-image library (http://scikit-image.org/docs/stable/api/skimage.segmentation.html).

:param inputImg: input image file.
:param markersImg: input markers image file.
:param outputImg: output image file.
:param gdalformat: output image file format.
:param tmpDIR: temp DIR used to output PCA files
:param calcStats: calculate image pixel statistics, histogram and image pyramids - note if you are not using a KEA file then the format needs to support RATs for this option as histogram and colour table are written to RAT.
:param usePCA: if there are not 1 or 3 image bands in the input file then you can use PCA to reduce the number of image bands.
:param nPCABands: the number of principle components outputs from the PCA - needs to be either 1 or 3.
:param compactness: scikit-image Watershed parameter: 'Use compact watershed with given compactness parameter. Higher values result in more regularly-shaped watershed basins; Peer Neubert & Peter Protzel (2014). Compact Watershed and Preemptive SLIC: On Improving Trade-offs of Superpixel Segmentation Algorithms. ICPR 2014'
:param watershed_line: scikit-image Watershed parameter: 'If watershed_line is True, a one-pixel wide line separates the regions obtained by the watershed algorithm. The line has the label 0.'

"""
    # Create output image
    rsgislib.imageutils.createCopyImage(inputImg, outputImg, 1, 0, gdalformat, rsgislib.TYPE_32UINT)
            
    
    gdalMarkersDS = gdal.Open(markersImg, gdal.GA_ReadOnly)
    if gdalMarkersDS.RasterCount != 1:
        gdalMarkersDS = None
        raise Exception('Markers image should only have one image band.')
    
    gdalMarkerBand = gdalMarkersDS.GetRasterBand(1)
    markerArr = gdalMarkerBand.ReadAsArray()
    
    gdalDS = gdal.Open(inputImg, gdal.GA_ReadOnly)
    pcaImg = ''
    if usePCA and (nPCABands == 1) and (gdalDS.RasterCount > 1):
        # Perform PCA
        gdalDS = None
        rsgisUtils = rsgislib.RSGISPyUtils()
        inImgBaseName = os.path.splitext(os.path.basename(inputImg))[0]
        pcaImg = os.path.join(tmpDIR, inImgBaseName+'_pca.kea')
        pcaEigenValsFile = os.path.join(tmpDIR, inImgBaseName+'_eigen.mtxt')
        inDType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImg)
        rsgislib.imagecalc.performImagePCA(inputImg, pcaImg, pcaEigenValsFile, 1, pcaPxlSample, 'KEA', inDType, noDataVal, False)
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif usePCA and (nPCABands == 3) and (gdalDS.RasterCount > 3):
        # Perform PCA
        gdalDS = None
        rsgisUtils = rsgislib.RSGISPyUtils()
        inImgBaseName = os.path.splitext(os.path.basename(inputImg))[0]
        pcaImg = os.path.join(tmpDIR, inImgBaseName+'_pca.kea')
        pcaEigenValsFile = os.path.join(tmpDIR, inImgBaseName+'_eigen.mtxt')
        inDType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImg)
        rsgislib.imagecalc.performImagePCA(inputImg, pcaImg, pcaEigenValsFile, 3, pcaPxlSample, 'KEA', inDType, noDataVal, False)
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif usePCA:
        gdalDS = None
        raise Exception('In the number of bands given the number of principle components.')    
    
    if not ((gdalDS.RasterCount == 1) or (gdalDS.RasterCount == 3)):
        gdalDS = None
        raise Exception('Input image should have either a single band or three (RGB). Consider performing PCA to reduce.')
    
    pxlValsArr = []
    if gdalDS.RasterCount > 1:
        for nBand in numpy.arange(gdalDS.RasterCount):
            gdalBand = gdalDS.GetRasterBand(int(nBand+1))
            pxlValsArr.append(gdalBand.ReadAsArray())
        pxlVals = numpy.stack(pxlValsArr, axis=-1)
    else:
        gdalBand = gdalDS.GetRasterBand(1)
        pxlVals = gdalBand.ReadAsArray()
    
    segResult = skimage.segmentation.watershed(pxlVals, markerArr, compactness=compactness, watershed_line=watershed_line)
    if not watershed_line:
        segResult = segResult + 1
    
    gdalOutDS = gdal.Open(outputImg, gdal.GA_Update)
    gdalOutBand = gdalOutDS.GetRasterBand(1)
    gdalOutBand.WriteArray(segResult)
    gdalOutDS = None
    gdalDS = None
    
    if calcStats:
        rsgislib.rastergis.populateStats(clumps=outputImg, addclrtab=True, calcpyramids=True, ignorezero=True)


def performRandomWalkerSegmentation(inputImg, markersImg, outputImg, gdalformat='KEA', noDataVal=0, tmpDIR='./tmp', calcStats=True, usePCA=False, nPCABands=3, pcaPxlSample=100, beta=130, mode='bf', tol=0.001, spacing=None):
    """
A function to perform the random walker segmentation algorithm from the
scikit-image library (http://scikit-image.org/docs/stable/api/skimage.segmentation.html).

:param inputImg: input image file.
:param markersImg: input markers image file - markers must be uniquely numbered.
:param outputImg: output image file.
:param gdalformat: output image file format.
:param tmpDIR: temp DIR used to output PCA files
:param calcStats: calculate image pixel statistics, histogram and image pyramids - note if you are not using a KEA file then the format needs to support RATs for this option as histogram and colour table are written to RAT.
:param usePCA: if there are not 1 or 3 image bands in the input file then you can use PCA to reduce the number of image bands.
:param nPCABands: the number of principle components outputs from the PCA - needs to be either 1 or 3.
:param beta: scikit-image random_walker parameter: 'Penalization coefficient for the random walker motion (the greater beta, the more difficult the diffusion).'
:param mode: scikit-image random_walker parameter: 'Mode for solving the linear system in the random walker algorithm. Available options {'cg_mg', 'cg', 'bf'}.'
        * 'bf' (brute force): an LU factorization of the Laplacian is computed. This is fast for small images (<1024x1024), but very slow and memory-intensive for large images (e.g., 3-D volumes).
        * 'cg' (conjugate gradient): the linear system is solved iteratively using the Conjugate Gradient method from scipy.sparse.linalg. This is less memory-consuming than the brute force method for large images, but it is quite slow.
        * 'cg_mg' (conjugate gradient with multigrid preconditioner): a preconditioner is computed using a multigrid solver, then the solution is computed with the Conjugate Gradient method. This mode requires that the pyamg module (http://pyamg.org/) is installed. For images of size > 512x512, this is the recommended (fastest) mode.
:param tol: scikit-image random_walker parameter: 'tolerance to achieve when solving the linear system, in cg’ and ‘cg_mg’ modes.'
:param spacing: scikit-image random_walker parameter: 'Spacing between voxels in each spatial dimension. If None, then the spacing between pixels/voxels in each dimension is assumed 1.'

"""
    # Create output image
    rsgislib.imageutils.createCopyImage(inputImg, outputImg, 1, 0, gdalformat, rsgislib.TYPE_32UINT)
    
    gdalMarkersDS = gdal.Open(markersImg, gdal.GA_ReadOnly)
    if gdalMarkersDS.RasterCount != 1:
        gdalMarkersDS = None
        raise Exception('Markers image should only have one image band.')
    
    gdalMarkerBand = gdalMarkersDS.GetRasterBand(1)
    markerArr = gdalMarkerBand.ReadAsArray()
    
    
    gdalDS = gdal.Open(inputImg, gdal.GA_ReadOnly)
    pcaImg = ''
    if usePCA and (nPCABands == 1) and (gdalDS.RasterCount > 1):
        # Perform PCA
        gdalDS = None
        rsgisUtils = rsgislib.RSGISPyUtils()
        inImgBaseName = os.path.splitext(os.path.basename(inputImg))[0]
        pcaImg = os.path.join(tmpDIR, inImgBaseName+'_pca.kea')
        pcaEigenValsFile = os.path.join(tmpDIR, inImgBaseName+'_eigen.mtxt')
        inDType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImg)
        rsgislib.imagecalc.performImagePCA(inputImg, pcaImg, pcaEigenValsFile, 1, pcaPxlSample, 'KEA', inDType, noDataVal, False)
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif usePCA and (nPCABands == 3) and (gdalDS.RasterCount > 3):
        # Perform PCA
        gdalDS = None
        rsgisUtils = rsgislib.RSGISPyUtils()
        inImgBaseName = os.path.splitext(os.path.basename(inputImg))[0]
        pcaImg = os.path.join(tmpDIR, inImgBaseName+'_pca.kea')
        pcaEigenValsFile = os.path.join(tmpDIR, inImgBaseName+'_eigen.mtxt')
        inDType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImg)
        rsgislib.imagecalc.performImagePCA(inputImg, pcaImg, pcaEigenValsFile, 3, pcaPxlSample, 'KEA', inDType, noDataVal, False)
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif usePCA:
        gdalDS = None
        raise Exception('In the number of bands given the number of principle components.')  
    
    if not ((gdalDS.RasterCount == 1) or (gdalDS.RasterCount == 3)):
        gdalDS = None
        raise Exception('Input image should have either a single band or three (RGB). Consider performing PCA to reduce.')
    
    multichannel = True
    pxlValsArr = []
    if gdalDS.RasterCount > 1:
        for nBand in numpy.arange(gdalDS.RasterCount):
            gdalBand = gdalDS.GetRasterBand(int(nBand+1))
            pxlValsArr.append(gdalBand.ReadAsArray())
        pxlVals = numpy.stack(pxlValsArr, axis=-1)
        multichannel = True
    else:
        gdalBand = gdalDS.GetRasterBand(1)
        pxlVals = gdalBand.ReadAsArray()
        multichannel = False
    
    segResult = skimage.segmentation.random_walker(pxlVals, markerArr, beta=beta, mode=mode, tol=tol, multichannel=multichannel, spacing=spacing)
    segResult = segResult + 1
    
    gdalOutDS = gdal.Open(outputImg, gdal.GA_Update)
    gdalOutBand = gdalOutDS.GetRasterBand(1)
    gdalOutBand.WriteArray(segResult)
    gdalOutDS = None
    gdalDS = None
    
    if calcStats:
        rsgislib.rastergis.populateStats(clumps=outputImg, addclrtab=True, calcpyramids=True, ignorezero=True)




"""
The tools module contains some useful tools and utilities for remote sensing work.
"""
from .plotting import *
from .projection import *
from .filemanipulation import *
from .visualisation import *
from .utm import *


#!/usr/bin/env python

#############################################
# rsgisimginfo.py
#
#  Copyright 2019 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
# Purpose:  Get information on vector files/layers. 
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 02/01/2019
# Version: 1.0
# 
#############################################

import argparse
import osgeo.gdal as gdal
from rios import rat

def listRATColumns(imgFile, band=1):
    dataset = gdal.Open(imgFile, gdal.GA_ReadOnly)
    nBands = dataset.RasterCount
    if (band == 0) or (band > nBands):
        raise Exception("Specified band is within the image.")
    bandObj = dataset.GetRasterBand(band)
    colNames = rat.getColumnNamesFromBand(bandObj)
    i = 1
    for col in colNames:
        print("{0}: {1}".format(i, col))
        i = i + 1

if __name__ == "__main__":

    # Read in config file
    parser = argparse.ArgumentParser(description="Obtain image info.")
    
    parser.add_argument("--ratcols", 
                        action='store_true', 
                        default=False, 
                        help="List the columns in a RAT.")
    parser.add_argument("-i", "--imgfile",
                        type=str,
                        required=True,
                        help="image file")
    parser.add_argument("-b", "--imgband",
                        type=int,
                        default=1,
                        help="image band")
    args = parser.parse_args()
    
    if args.ratcols:
        listRATColumns(args.imgfile, args.imgband)

